import os
from typing import Dict
from syngen.ml.metrics import UnivariateMetric, BaseTest
from datetime import datetime
import shutil

import jinja2
import pandas as pd

from syngen.ml.metrics.utils import transform_to_base64


class SampleAccuracyTest(BaseTest):
    def __init__(
        self,
        original: pd.DataFrame,
        sampled: pd.DataFrame,
        paths: dict,
        table_name: str,
        train_config: Dict,
    ):
        super().__init__(original, sampled, paths, table_name, train_config)
        self.reports_path = f"{self.paths['reports_path']}/sample_accuracy"

    def __get_univariate_metric(self):
        """
        Do preparation work before creating the report
        """
        self._prepare_dir()
        return UnivariateMetric(
            self.original,
            self.synthetic,
            True,
            self.reports_path
        )

    def __remove_artifacts(self):
        """
        Remove artifacts after creating Sample report
        """
        shutil.rmtree(self.reports_path)

    def report(self, **kwargs):
        univariate = self.__get_univariate_metric()
        uni_images = univariate.calculate_all(
            kwargs["cont_columns"], kwargs["categorical_columns"], kwargs["date_columns"]
        )

        # Generate html report
        with open(
            f"{os.path.dirname(os.path.realpath(__file__))}/sample_report_template.html"
        ) as file_:
            template = jinja2.Template(file_.read())

        uni_images = {
            title: transform_to_base64(path)
            for title, path in uni_images.items()
            if "word_count" not in title
        }

        config = {k: v for k, v in self.config.items() if k != "print_report"}

        html = template.render(
            uni_imgs=uni_images,
            table_name=self.table_name,
            config=config,
            time=datetime.now().strftime("%H:%M:%S %d/%m/%Y"),
        )

        with open(
            f"{self.paths['reports_path']}/sample_accuracy_report.html",
            "w",
            encoding="utf-8",
        ) as f:
            f.write(html)

        self._remove_artifacts()



import os
import traceback
from typing import Optional, List

import click
from loguru import logger
os.environ["TF_CPP_MIN_LOG_LEVEL"] = "2"

from syngen.ml.worker import Worker
from syngen.ml.utils import (
    setup_log_process,
    check_if_logs_available,
    validate_parameter_reports
)
from syngen.ml.validation_schema import ReportTypes


validate_reports = validate_parameter_reports(
    report_types=ReportTypes().train_report_types,
    full_list=ReportTypes().full_list_of_train_report_types
)


@click.command()
@click.option(
    "--metadata_path",
    type=str,
    default=None,
    help="Path to the metadata file"
)
@click.option(
    "--source",
    type=str,
    default=None,
    help="Path to the table that you want to use as a reference",
)
@click.option(
    "--table_name",
    type=str,
    default=None,
    help="Arbitrary string to name the directories",
)
@click.option(
    "--epochs",
    default=10,
    type=click.IntRange(1),
    help="Number of trained epochs. If absent, it's defaulted to 10",
)
@click.option(
    "--drop_null",
    default=False,
    type=click.BOOL,
    help="Flag which set whether to drop rows with at least one missing value. "
    "If absent, it's defaulted to False",
)
@click.option(
    "--row_limit",
    default=None,
    type=click.IntRange(1),
    help="Number of rows to train over. A number less than the original table "
         "length will randomly subset the specified rows number",
)
@click.option(
    "--reports",
    default=("none",),
    type=click.UNPROCESSED,
    multiple=True,
    callback=validate_reports,
    help="Controls the generation of quality reports. "
    "Might require significant time for big generated tables (>1000 rows). "
    "If set to 'sample', generates a sample report. "
    "If set to 'accuracy', generates an accuracy report. "
    "If set to 'metrics_only', outputs the metrics information only to standard output "
    "without generation of a report. "
    "If set to 'all', generates both accuracy and sample report. "
    "If it's absent or set to 'none', no reports are generated.",
)
@click.option(
    "--log_level",
    default="INFO",
    type=click.Choice(["TRACE", "DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]),
    help="Set the logging level which will be used in the process. "
         "If absent, it's defaulted to 'INFO'",
)
@click.option(
    "--batch_size",
    default=32,
    type=click.IntRange(1),
    help="Number of rows that goes in one batch. "
         "This parameter can help to control memory consumption.",
)
def launch_train(
    metadata_path: Optional[str],
    source: Optional[str],
    table_name: Optional[str],
    epochs: int,
    drop_null: bool,
    row_limit: Optional[int],
    reports: List[str],
    log_level: str,
    batch_size: int = 32,
):
    """
    Launch the work of training process

    Parameters
    ----------
    metadata_path
    source
    table_name
    epochs
    drop_null
    row_limit
    reports
    log_level
    batch_size
    -------

    """
    setup_log_process(
        type_of_process="train",
        log_level=log_level,
        table_name=table_name,
        metadata_path=metadata_path
    )
    if not metadata_path and not source and not table_name:
        raise AttributeError(
            "It seems that the information of 'metadata_path' or 'table_name' "
            "and 'source' is absent. Please provide either the information of "
            "'metadata_path' or the information of 'source' and 'table_name'"
        )
    elif metadata_path and table_name and source:
        logger.warning(
            "The information of 'metadata_path' was provided. "
            "In this case the information of 'table_name' and 'source' will be ignored"
        )
        table_name = None
    elif metadata_path and source:
        logger.warning(
            "The information of 'metadata_path' was provided. "
            "In this case the information of 'source' will be ignored"
        )
    elif metadata_path and table_name:
        logger.warning(
            "The information of 'metadata_path' was provided. "
            "In this case the information of 'table_name' will be ignored"
        )
        table_name = None
    elif source and not table_name:
        raise AttributeError(
            "It seems that the information of 'metadata_path' or 'table_name' is absent. "
            "Please provide either the information of 'metadata_path' or "
            "the information of 'source' and 'table_name'"
        )
    elif table_name and not source:
        raise AttributeError(
            "It seems that the information of 'metadata_path' or 'source' is absent. "
            "Please provide either the information of 'metadata_path' or "
            "the information of 'source' and 'table_name'"
        )
    logger.info(
        "The training process will be executed according to the information mentioned "
        "in 'train_settings' in the metadata file. If appropriate information is absent "
        "from the metadata file, then the values of parameters sent through CLI will be used. "
        "Otherwise, the values of parameters will be defaulted"
    )
    settings = {
        "source": source,
        "epochs": epochs,
        "drop_null": drop_null,
        "row_limit": row_limit,
        "batch_size": batch_size,
        "reports": reports,
    }

    worker = Worker(
        table_name=table_name,
        metadata_path=metadata_path,
        settings=settings,
        log_level=log_level,
        type_of_process="train",
    )

    worker.launch_train()


if __name__ == "__main__":
    try:
        launch_train()
    except Exception as e:
        log_file = os.getenv("SUCCESS_LOG_FILE")
        if not os.path.exists(log_file):
            logger.error(
                f"Training failed on running stage. "
                f"The traceback of the error - {traceback.format_exc()}"
            )
        raise e
    finally:
        check_if_logs_available()



from typing import Dict, List, Literal
import os
from dataclasses import dataclass, field
import json
from collections import defaultdict

from slugify import slugify
from loguru import logger
from syngen.ml.data_loaders import MetadataLoader, DataLoader
from syngen.ml.validation_schema import ValidationSchema, ReportTypes
from syngen.ml.utils import ValidationError


@dataclass
class Validator:
    """
    Class for validating the metadata
    """

    metadata: Dict
    metadata_path: str
    type_of_process: Literal["train", "infer"]
    validation_source: bool = True
    type_of_fk_keys = ["FK"]
    infer_report_types: List[str] = field(
        default_factory=lambda: ReportTypes().infer_report_types
    )
    merged_metadata: Dict = field(default_factory=dict)
    mapping: Dict = field(default_factory=dict)
    existed_columns_mapping: Dict = field(default_factory=dict)
    errors = defaultdict(defaultdict)

    @staticmethod
    def _check_encryption_process():
        """
        Check if the encryption process is enabled
        """
        fernet_key = os.getenv("FERNET_KEY")
        state = "is" if fernet_key else "isn't"
        process_status = "enabled" if fernet_key else "disabled"
        message = (
            f"As the environment variable 'FERNET_KEY' {state} set, "
            f"the encryption and decryption process is {process_status}."
        )
        logger.warning(message)

    def _launch_validation_of_schema(self):
        """
        Launch the validation of the schema of the metadata
        """
        return ValidationSchema(
            metadata=self.metadata,
            metadata_path=self.metadata_path,
            validation_source=self.validation_source,
            process=self.type_of_process
        ).validate_schema()

    def _define_mapping(self):
        """
        Define the mapping contained the information of the foreign keys
        defined in the metadata
        """
        for table_name, table_metadata in self.metadata.items():
            if table_name == "global":
                continue
            metadata_keys = (
                table_metadata.get("keys")
                if "keys" in table_metadata and table_metadata.get("keys") is not None
                else {}
            )
            for key_name, key_data in metadata_keys.items():
                if key_data["type"] not in self.type_of_fk_keys:
                    continue

                self.mapping[key_name] = {
                    "parent_table": key_data["references"]["table"],
                    "parent_columns": key_data["references"]["columns"],
                }

    def _check_conditions(self, metadata: Dict) -> bool:
        """
        Check conditions whether to launch validation or not
        """
        reports = metadata.get("train_settings", {}).get("reports", [])
        return (
            self.type_of_process == "infer"
            or (
                self.type_of_process == "train" and
                any([item in self.infer_report_types for item in reports])
            )
        )

    def _validate_metadata(self, table_name: str):
        """
        Validate the metadata
        """
        metadata_of_the_table = self.metadata[table_name]
        table_keys = metadata_of_the_table.get("keys", {})
        for key, config in table_keys.items():
            if config["type"] not in self.type_of_fk_keys:
                continue
            self._validate_referential_integrity(
                fk_name=key,
                fk_config=config,
                parent_config=self.merged_metadata[self.mapping[key]["parent_table"]],
            )
            parent_table = self.mapping[key]["parent_table"]
            if parent_table not in self.metadata:
                if self._check_conditions(metadata_of_the_table):
                    self._check_completion_of_training(parent_table)
                    self._check_existence_of_generated_data(parent_table)
                elif self.type_of_process == "train":
                    self._check_completion_of_training(parent_table)
            else:
                continue

    def _validate_referential_integrity(self, fk_name: str, fk_config: Dict, parent_config: Dict):
        """
        Validate whether the columns related to the primary key are the same as
        the referenced columns of the foreign key
        """
        keys = parent_config.get("keys", {})
        if not keys or all(config["type"] not in ["PK", "UQ"] for config in keys.values()):
            message = (
                "The information about columns of the primary or unique key "
                f"associated with the columns of the '{fk_config['type']}' key - '{fk_name}' "
                "wasn't found in the metadata of the parent table"
            )
            self.errors["validate referential integrity"][fk_name] = message
            return

        result = any(
            [
                config["columns"] == fk_config["references"]["columns"]
                for config in parent_config.get("keys", {}).values()
                if config["type"] in ["PK", "UQ"]
            ]
        )
        if result is False:
            message = (
                f"The columns of primary or unique key associated with the columns of "
                f"the '{fk_config['type']}' key - '{fk_name}' aren't the same"
            )
            self.errors["validate referential integrity"][fk_name] = message

    def _check_existence_of_generated_data(self, parent_table: str):
        """
        Check if the generated data of the certain parent table exists.
        The generated data is created after the successful execution of the inference process
        of the certain table.
        """
        destination = (
            self.merged_metadata[parent_table].get("infer_settings", {}).get("destination")
        )
        if destination is None:
            destination = (
                f"model_artifacts/tmp_store/{slugify(parent_table)}/"
                f"merged_infer_{slugify(parent_table)}.csv"
            )
        if not DataLoader(destination).has_existed_path:
            message = (
                f"The generated data of the table - '{parent_table}' hasn't been generated. "
                f"Please, generate the data related to the table '{parent_table}' first"
            )
            self.errors["check existence of the generated data"][parent_table] = message

    def _check_existence_of_source(self, table_name: str):
        """
        Check if the source of the certain table exists
        """
        if not DataLoader(
            self.merged_metadata[table_name]["train_settings"]["source"]
        ).has_existed_path:
            message = (
                f"It seems that the path to the source of the table - '{table_name}' "
                f"isn't correct. Please, check the path to the source of the table - "
                f"'{table_name}'"
            )
            self.errors["check existence of the source"][table_name] = message

    def _check_existence_of_destination(self, table_name: str):
        """
        Check if the destination of the certain table exists
        """
        destination = self.merged_metadata[table_name].get("infer_settings", {}).get("destination")
        if destination is None:
            logger.warning(
                f"As the destination path wasn't specified for the table - "
                f"'{table_name}', the synthetic data will be stored "
                f"at the default path - './model_artifacts/tmp_store/{slugify(table_name)}/"
                f"merged_infer_{slugify(table_name)}.csv'"
            )
        if destination is not None and not DataLoader(destination).has_existed_destination:
            message = (
                f"It seems that the directory path for storing the generated data of table "
                f"'{table_name}' isn't correct. Please, verify the destination path"
            )
            self.errors["check existence of the destination"][table_name] = message

    def _check_completion_of_training(self, table_name: str):
        """
        Check if the training process of a specific table has been completed.

        Args:
        table_name (str): The name of the table to check.

        Raises:
        FileNotFoundError: If the success file does not exist.
        ValueError: If the content of the success file does not indicate success.
        """
        path_to_success_file = f"model_artifacts/resources/{slugify(table_name)}/message.success"
        error_message = (
            f"The training of the table - '{table_name}' hasn't been completed. "
            "Please, retrain the table."
        )

        if os.path.exists(path_to_success_file):
            with open(path_to_success_file, "r") as file:
                content = file.read().strip()

        if not os.path.exists(path_to_success_file) or content != "SUCCESS":
            self.errors["check completion of the training process"][table_name] = error_message

    def _check_merged_metadata(self, parent_table: str):
        if parent_table not in self.merged_metadata:
            message = (
                f"The metadata of the parent table - '{parent_table}' hasn't been found. "
                f"Please, check whether the information of the parent table - '{parent_table}' "
                f"exists in the current metadata file or in the metadata files stored in "
                f"'model_artifacts/metadata' directory"
            )
            logger.error(message)
            raise ValidationError(message)

    def _merge_metadata(self):
        """
        Find the parent metadata contained the parent table
        in the metadata files stored in 'model_artifacts/metadata' directory,
        and merge it with the metadata of the child table
        """
        self.merged_metadata = self.metadata.copy()
        for key_name, config in self.mapping.items():
            parent_table = config.get("parent_table")
            if parent_table in self.metadata:
                continue
            path_to_metadata_storage = "model_artifacts/metadata"
            for file in os.listdir(path_to_metadata_storage):
                path_to_metadata_file = os.path.join(path_to_metadata_storage, file)
                metadata = MetadataLoader(path_to_metadata_file).load_data()
                if parent_table not in metadata:
                    continue
                self._launch_validation_of_schema()
                self.merged_metadata.update(metadata)
                logger.info(
                    f"The metadata located in the path - '{path_to_metadata_storage}' "
                    f"has been merged with the current metadata as it contains "
                    f"the information of the parent table - '{parent_table}'"
                )
            self._check_merged_metadata(parent_table)

    def _check_existence_of_key_columns(self, table_name):
        """
        Check whether the key columns exist in the table
        """
        metadata_of_table = self.merged_metadata[table_name]
        existed_columns = self.existed_columns_mapping[table_name]
        for key, config_of_key in metadata_of_table.get("keys", {}).items():
            if all([column in existed_columns for column in config_of_key["columns"]]):
                continue
            else:
                non_existed_columns = [
                    f"{col!r}"
                    for col in set(config_of_key["columns"]).difference(set(existed_columns))
                ]
                message = (
                    f"The columns of the {config_of_key['type']} key '{key}' - "
                    f"{', '.join(non_existed_columns)} "
                    f"don't exist in the table - '{table_name}'"
                )
                self.errors["check existence of the key columns in 'columns'"][key] = message

    def _check_existence_of_referenced_columns(self, table_name: str):
        """
        Check whether the key columns exist in the referenced table
        """
        metadata_of_table = self.merged_metadata[table_name]
        for key, config_of_key in metadata_of_table.get("keys", {}).items():
            if config_of_key["type"] in self.type_of_fk_keys:
                referenced_table = config_of_key["references"]["table"]
                referenced_columns = config_of_key["references"]["columns"]
                existed_columns = self.existed_columns_mapping.get(referenced_table, {})
                non_existed_columns = [
                    f"{col!r}"
                    for col in referenced_columns
                    if col not in existed_columns
                ]
                if non_existed_columns:
                    message = (
                        f"The 'references.columns' of the {config_of_key['type']} '{key}' - "
                        f"{', '.join(non_existed_columns)} "
                        f"don't exist in the referenced table - '{referenced_table}'"
                    )
                    self.errors[
                        "check existence of the key columns in 'references.columns'"
                    ][key] = message

    def _fetch_existed_columns(self, table_name: str) -> List[str]:
        """
        Fetch the list of the columns of the source table
        """
        metadata_of_table = self.merged_metadata[table_name]
        format_settings = metadata_of_table.get("format", {})
        path_to_source = self._fetch_path_to_source(table_name)
        return DataLoader(path_to_source).get_columns(**format_settings)

    def _gather_existed_columns(self, table_name: str):
        """
        Fetch the list of the existed columns of the table
        and put it into the mapping of existed columns
        """
        existed_columns = self._fetch_existed_columns(table_name)
        self.existed_columns_mapping[table_name] = existed_columns

    def _preprocess_metadata(self):
        """
        Preprocess the metadata, set the metadata and the merged metadata
        """
        self._launch_validation_of_schema()
        self._define_mapping()
        self._merge_metadata()
        self.merged_metadata.pop("global", None)
        self.metadata.pop("global", None)

    def _fetch_path_to_source(self, table_name):
        """
        Fetch the path to the source of the certain table
        """
        return self.merged_metadata[table_name]["train_settings"]["source"]

    def _launch_validation(self):
        """
        Launch the validation process
        """
        if self.type_of_process == "train" and self.validation_source:
            for table_name in self.merged_metadata.keys():
                self._gather_existed_columns(table_name)

        for table_name in self.metadata.keys():
            if self.type_of_process == "train" and self.validation_source:
                self._check_existence_of_source(table_name)
                self._check_existence_of_key_columns(table_name)
                self._check_existence_of_referenced_columns(table_name)
            elif self.type_of_process == "infer":
                self._check_completion_of_training(table_name)
                self._check_existence_of_destination(table_name)

        for table_name in self.metadata.keys():
            self._validate_metadata(table_name)

    def _collect_errors(self):
        """
        Collect the errors found during the validation process
        """
        error_logs = []
        for section, errors_details in self.errors.items():
            error_log = f'"{section}": {json.dumps(errors_details, indent=4)}'
            error_logs.append(error_log)
        if self.errors:
            message = (
                f"The validation of the metadata has been failed. "
                f"The error(s) found in - \n{', '.join(error_logs)}"
            )
            logger.error(message)
            raise ValidationError(message)
        logger.info("The validation of the metadata has been passed successfully")

    def run(self):
        """
        Run the validation process
        """
        self._check_encryption_process()
        self._preprocess_metadata()
        self._launch_validation()
        self._collect_errors()



from dataclasses import dataclass, field
from typing import Optional, Dict, Tuple, Set, List, Callable, Literal
import os
from copy import deepcopy
import shutil
from datetime import datetime

import pandas as pd
from loguru import logger
from slugify import slugify

from syngen.ml.data_loaders import DataLoader, DataFrameFetcher
from syngen.ml.utils import slugify_attribute, fetch_unique_root
from syngen.ml.convertor import CSVConvertor


@dataclass
class TrainConfig:
    """
    The configuration class to set up the work of train process
    """

    source: Optional[str]
    epochs: int
    drop_null: bool
    row_limit: Optional[int]
    table_name: Optional[str]
    metadata: Dict
    metadata_path: Optional[str]
    reports: List[str]
    batch_size: int
    loader: Optional[Callable[[str], pd.DataFrame]]
    data: pd.DataFrame = field(init=False)
    initial_data_shape: Tuple[int, int] = field(init=False)
    paths: Dict = field(init=False)
    row_subset: int = field(init=False)
    schema: Dict = field(init=False)
    original_schema: Dict = field(init=False)
    slugify_table_name: str = field(init=False)
    columns: List = field(init=False)
    dropped_columns: Set = field(init=False)

    def __post_init__(self):
        self._set_paths()

    def __getstate__(self) -> Dict:
        """
        Return an updated config's instance
        """
        instance = self.__dict__.copy()
        attribute_keys_to_remove = ["loader", "data"]
        for attr_key in attribute_keys_to_remove:
            if attr_key in instance:
                del instance[attr_key]
        return instance

    def preprocess_data(self):
        self._extract_data()
        self._save_original_schema()
        self._remove_empty_columns()
        self._mark_removed_columns()
        self._prepare_data()
        self._check_reports()

    def to_dict(self) -> Dict:
        """
        Return the values of the settings of training process
        """
        return {
            "epochs": self.epochs,
            "drop_null": self.drop_null,
            "row_subset": self.row_subset,
            "batch_size": self.batch_size,
            "reports": self.reports
        }

    def _set_batch_size(self):
        """
        Set up "batch_size" for training process
        """
        self.batch_size = min(self.batch_size, self.row_subset)

    def _check_sample_report(self):
        """
        Check whether it is necessary to generate a certain report
        """
        if "sample" in self.reports and self.initial_data_shape[0] == self.row_subset:
            logger.warning(
                "The generation of the sample report is unnecessary and won't be produced "
                "as the source data and sampled data sizes are identical"
            )
            reports = deepcopy(self.reports)
            reports.remove("sample")
            self.reports = reports

    def _check_reports(self):
        """
        Check whether it is necessary to generate a certain report
        """
        self._check_sample_report()

    def _fetch_dataframe(self) -> Tuple[pd.DataFrame, Dict]:
        """
        Fetch the dataframe using the callback function
        """
        dataframe_fetcher = DataFrameFetcher(
            loader=self.loader,
            table_name=self.table_name
        )
        self.original_schema = dataframe_fetcher.original_schema
        return dataframe_fetcher.fetch_data()

    def _load_source(self) -> Tuple[pd.DataFrame, Dict]:
        """
        Return dataframe and schema of original data
        """
        if os.path.exists(self.paths["path_to_flatten_metadata"]):
            data_loader = DataLoader(self.paths["input_data_path"], sensitive=True)
            data, schema = data_loader.load_data()
            self.original_schema = data_loader.original_schema
            schema = CSVConvertor.schema
            return data, schema
        else:
            data_loader = DataLoader(self.source)
            self.original_schema = data_loader.original_schema
            if self.original_schema is not None:
                logger.trace(
                    f"The schema of the table - '{self.table_name}': {self.original_schema}"
                )
            return data_loader.load_data()

    def _remove_empty_columns(self):
        """
        Remove completely empty columns from dataframe
        """
        data_columns = set(self.data.columns)
        self.data = self.data.dropna(how="all", axis=1)

        self.dropped_columns = data_columns - set(self.data.columns)
        list_of_dropped_columns = [f"'{column}'" for column in self.dropped_columns]
        if len(list_of_dropped_columns) > 0:
            logger.info(f"Empty columns - {', '.join(list_of_dropped_columns)} were removed")

    def _mark_removed_columns(self):
        """
        Mark removed columns in the schema
        """
        if self.schema.get("format") == "CSV":
            self.schema["fields"] = dict()
            self.schema["fields"] = {column: "removed" for column in self.dropped_columns}
        else:
            for column, data_type in self.schema.get("fields", {}).items():
                if column not in self.data.columns:
                    self.schema["fields"][column] = "removed"

    def _check_if_data_is_empty(self):
        """
        Check if the provided data is empty
        """
        if self.data.shape[0] < 1:
            raise ValueError(
                f"The empty table was provided. Unable to train the table - '{self.table_name}'"
            )

    def _extract_data(self):
        """
        Extract data and schema necessary for training process
        """
        self.data, self.schema = self._load_source()
        self.initial_data_shape = self.data.shape
        self.columns = list(self.data.columns)
        self._check_if_data_is_empty()

    def _preprocess_data(self):
        """
        Preprocess data and set the parameter "row_subset" for training process
        """
        if self.loader:
            warning_message = (
                "parameter will be ignored because the retrieval of the data "
                "is handled by a callback function"
            )
            if self.drop_null:
                logger.warning(f"The 'drop_null' {warning_message}")
            if self.row_limit is not None:
                logger.warning(f"The 'row_limit' {warning_message}")
        else:
            if self.drop_null:
                if not self.data.dropna().empty:
                    initial_data = self.data
                    self.data = self.data.dropna()
                    if count_of_dropped_rows := initial_data.shape[0] - self.data.shape[0]:
                        logger.info(
                            f"As the parameter 'drop_null' set to 'True', "
                            f"{count_of_dropped_rows} rows of the table - '{self.table_name}' "
                            f"that have empty values have been dropped. "
                            f"The count of remained rows is {self.data.shape[0]}."
                        )
                else:
                    logger.warning(
                        "The specified 'drop_null' argument results in the empty dataframe, "
                        "so it will be ignored"
                    )

            if self.row_limit:
                self.row_subset = min(self.row_limit, len(self.data))

                self.data = self.data.sample(n=self.row_subset)

        if len(self.data) < 100:
            logger.warning(
                "The input table is too small to provide any meaningful results. "
                "Please consider 1) disable drop_null argument, 2) provide bigger table"
            )
        elif len(self.data) < 500:
            logger.warning(
                f"The amount of data is {len(self.data)} rows. It seems that it isn't enough "
                f"to supply high-quality results. To improve the quality of generated data "
                f"please consider any of the steps: 1) provide a bigger table, "
                f"2) disable drop_null argument"
            )

        logger.info(f"The subset of rows was set to {len(self.data)}")

        self.row_subset = len(self.data)
        self._set_batch_size()

    def _save_input_data(self):
        """
        Save the subset of the original data
        """
        DataLoader(self.paths["input_data_path"], sensitive=True).save_data(self.data)

    def _save_original_schema(self):
        """
        Save the schema of the original data
        """
        DataLoader(self.paths["original_schema_path"]).save_data(self.original_schema)

    def _prepare_data(self):
        """
        Preprocess and save the data necessary for the training process
        """
        self._preprocess_data()
        if not self.loader:
            self._save_input_data()

    @slugify_attribute(table_name="slugify_table_name")
    def _set_paths(self):
        """
        Create the paths which used in training process
        """
        losses_file_name = (
            f"losses_{self.table_name}_"
            f"{datetime.now().strftime('%Y-%m-%d %H:%M:%S')}"
        )
        self.paths = {
            "model_artifacts_path": "model_artifacts/",
            "resources_path": f"model_artifacts/resources/{self.slugify_table_name}/",
            "tmp_store_path": f"model_artifacts/tmp_store/{self.slugify_table_name}/",
            "source_path": self.source,
            "reports_path": f"model_artifacts/tmp_store/{self.slugify_table_name}/reports",
            "input_data_path": f"model_artifacts/tmp_store/{self.slugify_table_name}/"
                               f"input_data_{self.slugify_table_name}."
                               f"{'dat' if os.getenv('FERNET_KEY') else 'pkl'}",
            "state_path": f"model_artifacts/resources/{self.slugify_table_name}/vae/checkpoints",
            "train_config_pickle_path": f"model_artifacts/resources/{self.slugify_table_name}/vae/"
                                        f"checkpoints/train_config.pkl",
            "dataset_pickle_path": f"model_artifacts/resources/{self.slugify_table_name}/vae/"
                                   f"checkpoints/model_dataset.pkl",
            "fk_kde_path": f"model_artifacts/resources/{self.slugify_table_name}/vae/"
                           f"checkpoints/stat_keys/",
            "original_schema_path": f"model_artifacts/tmp_store/{self.slugify_table_name}/"
                                    f"original_schema_{self.slugify_table_name}.pkl",
            "path_to_merged_infer": f"model_artifacts/tmp_store/{self.slugify_table_name}/"
                                    f"merged_infer_{self.slugify_table_name}.csv",
            "no_ml_state_path":
                f"model_artifacts/resources/{self.slugify_table_name}/no_ml/checkpoints/",
            "path_to_flatten_metadata":
                f"model_artifacts/tmp_store/flatten_configs/"
                f"flatten_metadata_{fetch_unique_root(self.table_name, self.metadata_path)}.json",
            "losses_path": f"model_artifacts/tmp_store/losses/{slugify(losses_file_name)}.csv"
        }


@dataclass
class InferConfig:
    """
    The configuration class to set up the work of infer process
    """

    destination: Optional[str]
    metadata: Dict
    metadata_path: Optional[str]
    size: Optional[int]
    table_name: Optional[str]
    run_parallel: bool
    batch_size: Optional[int]
    random_seed: Optional[int]
    reports: List[str]
    both_keys: bool
    log_level: str
    loader: Optional[Callable[[str], pd.DataFrame]]
    type_of_process: Literal["train", "infer"]
    slugify_table_name: str = field(init=False)

    def __post_init__(self):
        self._set_paths()
        self._remove_artifacts()
        self._set_infer_parameters()

    def _set_infer_parameters(self):
        self._check_reports()
        self._set_up_size()
        self._set_up_batch_size()

    def _remove_reports(self):
        path_to_reports = self.paths["reports_path"]
        if os.path.exists(path_to_reports):
            shutil.rmtree(path_to_reports)
            logger.info(
                f"The reports generated in the previous run of an inference process "
                f"and located in the path - '{path_to_reports}' were removed"
            )

    def _remove_generated_data(self):
        default_path_to_synth_data = self.paths["default_path_to_merged_infer"]
        if os.path.exists(default_path_to_synth_data):
            os.remove(default_path_to_synth_data)
            logger.info(
                f"The synthetic data generated in the previous run of an inference process and "
                f"located in the path - '{default_path_to_synth_data}' was removed"
            )

    def _remove_artifacts(self):
        """
        Remove artifacts related to the previous generation process
        """
        self._remove_reports()
        self._remove_generated_data()

    def to_dict(self) -> Dict:
        """
        Return the values of the settings of inference process
        """
        return {
            "size": self.size,
            "run_parallel": self.run_parallel,
            "batch_size": self.batch_size,
            "random_seed": self.random_seed,
            "reports": self.reports,
        }

    def _check_required_artifacts(self):
        """
        Check whether required artifacts exists
        """
        data_loader = DataLoader(self.paths["input_data_path"], sensitive=True)
        if (
                self.reports
                and (
                    data_loader.has_existed_path is False
                    or self.loader is not None
                )
        ):
            self.reports = list()
            log_message = (
                f"It seems that the path to the sample of the original data for the table "
                f"'{self.table_name}' at '{self.paths['input_data_path']}' does not exist. "
                f"As a result, no reports for the table '{self.table_name}' will be generated. "
                f"The 'reports' parameter for the table '{self.table_name}' "
                f"has been set to 'none'."
            )
            logger.warning(log_message)

    def _check_reports(self):
        """
        Check whether it is possible to generate reports
        """
        self._check_required_artifacts()

    def _set_up_size(self):
        """
        Set up "size" of generated data
        """
        if self.size is None:
            data_loader = DataLoader(self.paths["input_data_path"], sensitive=True)
            data = pd.DataFrame()
            if data_loader.has_existed_path:
                data, schema = data_loader.load_data()
            elif self.loader:
                data, schema = DataFrameFetcher(
                    loader=self.loader,
                    table_name=self.table_name
                ).fetch_data()
            self.size = len(data)

    def _set_up_batch_size(self):
        """
        Set up "batch_size" of inference process
        """
        self.batch_size = (
            min(self.batch_size, self.size) if self.batch_size is not None else self.size
        )

    @slugify_attribute(table_name="slugify_table_name")
    def _set_paths(self):
        """
        Create the paths which used in inference process
        """
        dynamic_name = self.slugify_table_name[:-3] if self.both_keys else self.slugify_table_name
        self.paths = {
            "reports_path": f"model_artifacts/tmp_store/{dynamic_name}/reports",
            "input_data_path": f"model_artifacts/tmp_store/{self.slugify_table_name}/"
                               f"input_data_{self.slugify_table_name}."
                               f"{'dat' if os.getenv('FERNET_KEY') else 'pkl'}",
            "default_path_to_merged_infer": f"model_artifacts/tmp_store/{dynamic_name}/"
                                            f"merged_infer_{dynamic_name}.csv",
            "path_to_merged_infer": self.destination
            if self.destination is not None
            else f"model_artifacts/tmp_store/{dynamic_name}/merged_infer_{dynamic_name}.csv",
            "state_path": f"model_artifacts/resources/{dynamic_name}/vae/checkpoints",
            "train_config_pickle_path":
                f"model_artifacts/resources/{dynamic_name}/vae/checkpoints/train_config.pkl",
            "original_schema_path": f"model_artifacts/tmp_store/{self.slugify_table_name}/"
                                    f"original_schema_{self.slugify_table_name}.pkl",
            "tmp_store_path": f"model_artifacts/tmp_store/{dynamic_name}",
            "vae_resources_path": f"model_artifacts/resources/{dynamic_name}/vae/checkpoints/",
            "dataset_pickle_path":
                f"model_artifacts/resources/{dynamic_name}/vae/checkpoints/model_dataset.pkl",
            "fk_kde_path": f"model_artifacts/resources/{dynamic_name}/vae/checkpoints/stat_keys/",
            "path_to_no_ml":
                f"model_artifacts/resources/{dynamic_name}/no_ml/checkpoints/kde_params.pkl",
            "path_to_flatten_metadata":
                f"model_artifacts/tmp_store/flatten_configs/"
                f"flatten_metadata_{fetch_unique_root(self.table_name, self.metadata_path)}.json"
        }



import shutil
from abc import ABC, abstractmethod
from datetime import datetime
from typing import List, Dict, Optional
import os

import jinja2
import pandas as pd
import numpy as np
from loguru import logger

from syngen.ml.metrics import (
    UnivariateMetric,
    BivariateMetric,
    JensenShannonDistance,
    Correlations,
    Clustering,
    Utility,
)
from syngen.ml.metrics.utils import transform_to_base64
from syngen.ml.utils import fetch_config, ProgressBarHandler
from syngen.ml.mlflow_tracker import MlflowTracker
from syngen.ml.validation_schema import ReportTypes


class BaseTest(ABC):
    def __init__(
        self,
        original: pd.DataFrame,
        synthetic: pd.DataFrame,
        paths: dict,
        table_name: str,
        config: Dict,
    ):
        self.original = original
        self.synthetic = synthetic
        self.paths = paths
        self.table_name = table_name
        self.config = config
        self.plot_exists = any(
            [
                item in ReportTypes().full_list_of_infer_report_types
                for item
                in self.config.get("reports", [])
            ]
        )
        self.reports_path = str()

    @abstractmethod
    def report(
        self,
        cont_columns: List[str],
        categorical_columns: List[str],
        text_columns: List[str],
        date_columns: Optional[List[str]],
    ):
        pass

    def _prepare_dir(self):
        """
        Create the directory where images and reports should be stored
        """
        os.makedirs(self.reports_path, exist_ok=True)

    def _remove_artifacts(self):
        """
        Remove artifacts after creating the reports or fetching the metrics
        """
        shutil.rmtree(self.reports_path)
        base_dir = os.path.dirname(self.reports_path)
        if os.path.exists(base_dir) and not os.listdir(base_dir):
            os.rmdir(base_dir)

    def _log_report_to_mlflow(self, path):
        """
        Log report to mlflow
        """
        try:
            MlflowTracker().log_artifact(path)
        except Exception as error:
            logger.warning(
                f"Logging the report to mlflow has failed due to a permission error. "
                f"File path: '{path}', Error details: {error}.\n"
                f"The report will be saved locally in '{self.reports_path}'"
            )
            pass

    @staticmethod
    def update_progress_bar(message, delta=0):
        ProgressBarHandler().set_progress(
            progress=ProgressBarHandler().progress + delta, delta=None, message=message
        )

    def _get_cleaned_configs(self):
        """
        Get cleaned configs for the report
        """
        filtered_fields = ["reports"]
        train_config = {
            k: v
            for k, v in fetch_config(self.paths["train_config_pickle_path"])
            .to_dict()
            .items()
            if k not in filtered_fields
        }
        infer_config = {
            k: v
            for k, v in self.config.items()
            if k not in filtered_fields
        }
        return train_config, infer_config


class AccuracyTest(BaseTest):
    def __init__(
        self,
        original: pd.DataFrame,
        synthetic: pd.DataFrame,
        paths: dict,
        table_name: str,
        infer_config: Dict
    ):
        super().__init__(original, synthetic, paths, table_name, infer_config)
        self.reports_path = f"{self.paths['reports_path']}/accuracy"
        self.dataset_pickle_path = self.paths["dataset_pickle_path"]
        self.univariate = UnivariateMetric(
            self.original,
            self.synthetic,
            self.plot_exists,
            self.reports_path
        )
        self.bivariate = BivariateMetric(
            self.original,
            self.synthetic,
            self.plot_exists,
            self.reports_path,
            self.dataset_pickle_path
        )
        self.correlations = Correlations(
            self.original,
            self.synthetic,
            self.plot_exists,
            self.reports_path
        )
        self.clustering = Clustering(
            self.original,
            self.synthetic,
            self.plot_exists,
            self.reports_path
        )
        self.utility = Utility(
            self.original,
            self.synthetic,
            self.plot_exists,
            self.reports_path
        )
        self.acc = JensenShannonDistance(
            self.original,
            self.synthetic,
            self.plot_exists,
            self.reports_path
        )
        self._prepare_dir()

    def _fetch_metrics(self, **kwargs):
        """
        Fetch the main metrics
        """
        delta = ProgressBarHandler().delta / 6

        self.update_progress_bar("Generation of the accuracy heatmap...")
        self.acc.calculate_all(kwargs["categorical_columns"])
        acc_median = round(self.acc.calculate_heatmap_median(self.acc.heatmap), 4)
        logger.info(f"Median accuracy is {acc_median}")
        self.update_progress_bar("The accuracy heatmap has been generated", delta)

        uni_images = dict()
        bi_images = dict()

        if self.plot_exists:
            self.update_progress_bar("Generation of the univariate distributions...")
            uni_images = self.univariate.calculate_all(
                kwargs["cont_columns"], kwargs["categorical_columns"], kwargs["date_columns"]
            )
            self.update_progress_bar(
                "The univariate distributions have been generated", delta
            )

            self.update_progress_bar("Generation of the bivariate distributions...")
            bi_images = self.bivariate.calculate_all(
                kwargs["cont_columns"], kwargs["categorical_columns"], kwargs["date_columns"]
            )
            self.update_progress_bar(
                "The bivariate distributions have been generated", delta
            )

        self.update_progress_bar("Generation of the correlations heatmap...")
        corr_result = self.correlations.calculate_all(
            kwargs["categorical_columns"], kwargs["cont_columns"]
        )
        corr_result = round(int(corr_result) if corr_result == 0 else abs(corr_result), 4)
        logger.info(f"Median of differences of correlations is {corr_result}")
        self.update_progress_bar("The correlations heatmap has been generated", delta)

        self.update_progress_bar("Generation of the clustering metric...")
        clustering_result = self.clustering.calculate_all(
            kwargs["categorical_columns"], kwargs["cont_columns"]
        )
        clustering_result = round(clustering_result, 4) if clustering_result is not None else None
        logger.info(f"Mean clusters homogeneity is {clustering_result}")
        self.update_progress_bar("The clustering metric has been calculated", delta)

        self.update_progress_bar("Generation of the utility metric...")
        utility_result = self.utility.calculate_all(
            kwargs["categorical_columns"], kwargs["cont_columns"]
        )
        self.update_progress_bar("The utility metric has been calculated", delta)

        return (
            acc_median,
            corr_result,
            clustering_result,
            utility_result,
            uni_images,
            bi_images,
        )

    def _generate_report(
        self,
        acc_median,
        corr_result,
        clustering_result,
        utility_result,
        uni_images,
        bi_images,
    ):
        """
        Generate the report
        """
        with open(
            f"{os.path.dirname(os.path.realpath(__file__))}/accuracy_report.html"
        ) as file_:
            template = jinja2.Template(file_.read())

        uni_images = {
            title: transform_to_base64(path) for title, path in uni_images.items()
        }
        bi_images = {
            title: transform_to_base64(path) for title, path in bi_images.items()
        }

        train_config, infer_config = self._get_cleaned_configs()

        html = template.render(
            accuracy_value=acc_median,
            accuracy_heatmap=transform_to_base64(
                f"{self.reports_path}/accuracy_heatmap.svg"
            ),
            uni_imgs=uni_images,
            correlations_heatmap=transform_to_base64(
                f"{self.reports_path}/correlations_heatmap.svg"
            ),
            correlation_median=corr_result,
            clusters_barplot=transform_to_base64(
                f"{self.reports_path}/clusters_barplot.svg"
            ),
            clustering_value=clustering_result,
            bi_imgs=bi_images,
            utility_barplot=transform_to_base64(
                f"{self.reports_path}/utility_barplot.svg"
            ),
            utility_table=utility_result.to_html() if utility_result is not None else None,
            is_data_available=(
                False
                if utility_result is None or (utility_result is not None and utility_result.empty)
                else True
            ),
            table_name=self.table_name,
            training_config=train_config,
            inference_config=infer_config,
            time=datetime.now().strftime("%H:%M:%S %d/%m/%Y"),
            round=round,
        )

        path_to_accuracy_report = f"{self.paths['reports_path']}/accuracy_report.html"
        with open(
            path_to_accuracy_report, "w", encoding="utf-8"
        ) as f:
            f.write(html)
        self._log_report_to_mlflow(path_to_accuracy_report)

    def report(self, *args, **kwargs):
        metrics = self._fetch_metrics(**kwargs)
        (
            acc_median,
            corr_result,
            clustering_result,
            utility_result,
            uni_images,
            bi_images,
        ) = metrics
        MlflowTracker().log_metrics(
            {
                "Utility_avg": (
                    utility_result["Synth to orig ratio"].mean()
                    if utility_result is not None else None
                ),
                "Clustering": (
                    clustering_result
                    if clustering_result is not None else np.NaN
                ),
                "Accuracy": acc_median,
                "Correlation": corr_result,
            }
        )

        if self.plot_exists:
            self._generate_report(
                acc_median,
                corr_result,
                clustering_result,
                utility_result,
                uni_images,
                bi_images,
            )
        self._remove_artifacts()



import os
import toml
from streamlit.web import bootstrap


def start():
    dir_path = os.path.dirname(__file__)
    file_path = os.path.join(dir_path, "run.py")
    config_path = os.path.join(dir_path, ".streamlit", "config.toml")

    config = toml.load(config_path)

    config_options = {}
    for section in config:
        for setting, value in config[section].items():
            option_name = f"{section}_{setting}"
            config_options[option_name] = value

    bootstrap.load_config_options(config_options)
    bootstrap.run(file_path, False, [], flag_options={})


if __name__ == "__main__":
    start()



from typing import List
import base64

import pandas as pd
from sklearn.preprocessing import OrdinalEncoder
from loguru import logger


def encode_categorical_features(dfs: List[pd.DataFrame]) -> List[pd.DataFrame]:
    encoder = OrdinalEncoder()
    dfs = [df.fillna("?") for df in dfs]

    encoder.fit(pd.concat(dfs, axis=0))
    res_dfs = []
    for df in dfs:
        encoded_df = encoder.transform(df)
        res_dfs.append(encoded_df)
    return res_dfs


def series_count_words(x):
    return len(str(x).split())


def set_empty_string(value):
    if value is None:
        return ""
    return str(value)


def text_to_continuous(df, text_columns: List[str]):
    for col in text_columns:
        df[col + "_char_len"] = df[col].apply(set_empty_string, convert_dtype=False).apply(len)
        df[col + "_word_count"] = (
            df[col].apply(set_empty_string, convert_dtype=False).apply(series_count_words)
        )
    return df


def transform_to_base64(path):
    try:
        with open(path, "rb") as image_file:
            encoded_string = base64.b64encode(image_file.read())
        return "data:image/svg+xml;base64," + encoded_string.decode("utf-8")
    except FileNotFoundError:
        logger.warning(f"No file found at {path}")
        return ""



from abc import abstractmethod
from typing import (
    Dict,
    Tuple,
    Optional,
    Callable,
    Union,
    List
)
import itertools
from collections import defaultdict

import pandas as pd
import numpy as np
from loguru import logger

from syngen.ml.utils import (
    nan_labels_to_float,
    fetch_config,
    datetime_to_timestamp,
)
from syngen.ml.metrics import AccuracyTest, SampleAccuracyTest
from syngen.ml.data_loaders import DataLoader, DataFrameFetcher
from syngen.ml.metrics.utils import text_to_continuous
from syngen.ml.mlflow_tracker import MlflowTracker
from syngen.ml.utils import ProgressBarHandler


class Reporter:
    """
    Abstract class for reporters
    """

    def __init__(
        self,
        table_name: str,
        paths: Dict[str, str],
        config: Dict[str, str],
        loader: Optional[Callable[[str], pd.DataFrame]] = None
    ):
        self.table_name = table_name
        self.paths = paths
        self.config = config
        self.loader = loader
        self.dataset = None
        self.columns_nan_labels = dict()
        self.na_values = dict()

    def _fetch_dataframe(self) -> pd.DataFrame:
        """
        Fetch the data using the callback function
        """
        data, schema = DataFrameFetcher(
            loader=self.loader,
            table_name=self.table_name
        ).fetch_data()
        logger.warning(
            f"The original data of the table - '{self.table_name}' "
            "has been fetched using the callback function. "
            "The data may have been modified since the beginning of the training process."
        )
        return data

    def _extract_report_data(self) -> Tuple[pd.DataFrame, pd.DataFrame]:
        if self.loader:
            original = self._fetch_dataframe()
        else:
            original, schema = DataLoader(
                self.paths["input_data_path"],
                sensitive=True
            ).load_data()
        synthetic, schema = DataLoader(self.paths["path_to_merged_infer"]).load_data()
        return original, synthetic

    def fetch_data_types(self):
        self.dataset = fetch_config(self.paths["dataset_pickle_path"])
        self.columns_nan_labels = self.dataset.nan_labels_dict
        types = (
            self.dataset.str_columns,
            self.dataset.date_columns,
            self.dataset.int_columns,
            self.dataset.float_columns,
            self.dataset.binary_columns,
            self.dataset.categorical_columns,
            self.dataset.long_text_columns,
            self.dataset.email_columns,
        )

        # eliminate keys columns from the report
        keys_columns = (
            set(self.dataset.pk_columns) |
            set(self.dataset.fk_columns) |
            set(self.dataset.uq_columns)
        )
        types = tuple(columns - keys_columns for columns in types)

        return types

    @staticmethod
    def convert_to_timestamp(df, col_name, date_format, na_values):
        """
        Convert date column to timestamp
        """
        return [
            datetime_to_timestamp(d, date_format)
            if d not in na_values else np.NaN for d in df[col_name]
        ]

    def preprocess_data(self, original: pd.DataFrame, synthetic: pd.DataFrame):
        """
        Preprocess original and synthetic data.
        Return original data, synthetic data, float columns, integer columns, categorical columns
        without keys columns
        """
        types = self.fetch_data_types()
        missing_columns = set(original) - set(synthetic)
        for col in missing_columns:
            synthetic[col] = np.nan
        exclude_columns = self.dataset.uuid_columns
        for column in self.dataset.cast_to_integer:
            original[column] = pd.to_numeric(
                original[column], errors="coerce", downcast="integer"
            )
            synthetic[column] = pd.to_numeric(
                synthetic[column], errors="coerce", downcast="integer"
            )
        for column in self.dataset.cast_to_float:
            original[column] = pd.to_numeric(
                original[column], errors="coerce", downcast="float"
            )
            synthetic[column] = pd.to_numeric(
                synthetic[column], errors="coerce", downcast="float"
            )
        original = nan_labels_to_float(
            original, self.columns_nan_labels, exclude_columns, process="report"
        )
        synthetic = nan_labels_to_float(
            synthetic, self.columns_nan_labels, exclude_columns, process="report"
        )
        (
            str_columns,
            date_columns,
            int_columns,
            float_columns,
            binary_columns,
            categorical_columns,
            long_text_columns,
            email_columns,
        ) = types

        original = original[[col for col in original.columns if col in set().union(*types)]]
        synthetic = synthetic[[col for col in synthetic.columns if col in set().union(*types)]]
        na_values = self.dataset.format.get("na_values", [])
        for date_col, date_format in self.dataset.date_mapping.items():
            original[date_col] = self.convert_to_timestamp(
                original, date_col, date_format, na_values
            )
            synthetic[date_col] = self.convert_to_timestamp(
                synthetic, date_col, date_format, na_values
            )

        int_columns = date_columns | int_columns
        text_columns = str_columns | long_text_columns | email_columns
        original = text_to_continuous(original, text_columns).drop(text_columns, axis=1)
        synthetic = text_to_continuous(synthetic, text_columns).drop(text_columns, axis=1)

        for col in [i + "_word_count" for i in text_columns]:
            if original[col].nunique() < 50:  # ToDo check if we need this
                categorical_columns = categorical_columns | {col}
            else:
                int_columns = int_columns | {col}
        int_columns = int_columns | {i + "_char_len" for i in text_columns}

        categorical_columns = categorical_columns | binary_columns

        for col in categorical_columns:
            original[col] = original[col].astype(str)
            synthetic[col] = synthetic[col].astype(str)
        return (
            original,
            synthetic,
            float_columns,
            int_columns,
            categorical_columns,
            date_columns,
        )

    @abstractmethod
    def report(self, **kwargs):
        """
        Generate the report for certain test
        """
        pass


class Report:
    """
    Singleton metaclass for registration all needed reporters
    """

    _reporters: Dict[str, Union[Reporter, List]] = {}

    def __new__(cls):
        if not hasattr(cls, "instance"):
            cls.instance = super(Report, cls).__new__(cls)
        return cls.instance

    @classmethod
    def register_reporter(cls, table: str, reporter: Reporter):
        """
        Register all needed reporters
        """
        list_of_reporters = cls._reporters.get(table, [])
        list_of_reporters.append(reporter)
        cls._reporters[table] = list_of_reporters

    @classmethod
    def unregister_reporters(cls, table: str):
        """
        Unregister all reporters for a table
        """
        cls._reporters[table] = list()

    @classmethod
    def clear_report(cls):
        """
        Delete unnecessary reporters
        """
        cls._reporters.clear()

    @classmethod
    def _group_reporters(cls):
        """
        Group reporters by table name
        """
        list_of_reporters = itertools.chain.from_iterable(cls._reporters.values())
        sorted_reporters = sorted(list_of_reporters, key=lambda r: r.table_name)
        grouped_reporters = defaultdict(list)

        for reporter in sorted_reporters:
            grouped_reporters[reporter.table_name].append(reporter)

        return grouped_reporters

    @classmethod
    def generate_report(cls):
        """
        Generate all needed reports
        """
        grouped_reporters = cls._group_reporters()

        if grouped_reporters:
            logger.warning("The report(s) generation might be time-consuming")

        for table_name, reporters in grouped_reporters.items():
            cls._start_mlflow_run(table_name)
            delta = 0.25 / len(reporters)

            for reporter in reporters:
                cls._launch_reporter(reporter, delta)

            MlflowTracker().end_run()

    @staticmethod
    def _start_mlflow_run(table_name: str):
        MlflowTracker().start_run(
            run_name=f"{table_name}-REPORT",
            tags={"table_name": table_name, "process": "report"},
        )

    @classmethod
    def _launch_reporter(cls, reporter, delta: float):
        cls._log_and_update_progress(
            delta,
            f"The calculation of {reporter.__class__.report_type} metrics for the table - "
            f"'{reporter.table_name}' has started"
        )

        reporter.report()

        if (
                reporter.__class__.report_type == "accuracy"
                and "accuracy" not in reporter.config["reports"]
                and "metrics_only" in reporter.config["reports"]
        ):
            message = (
                f"The metrics for the table - '{reporter.table_name}' have been evaluated"
            )
        else:
            message = (
                f"The {reporter.__class__.report_type} report of the table - "
                f"'{reporter.table_name}' has been generated"
            )
        cls._log_and_update_progress(delta, message)

    @staticmethod
    def _log_and_update_progress(delta: float, message: str):
        ProgressBarHandler().set_progress(delta=delta, message=message)
        logger.info(message)
        ProgressBarHandler().set_progress(
            progress=ProgressBarHandler().progress + delta,
            delta=delta,
            message=message
        )

    @property
    def reporters(self) -> Dict[str, Reporter]:
        return self._reporters


class AccuracyReporter(Reporter):
    """
    Reporter for running an accuracy test
    """

    report_type = "accuracy"

    def report(self):
        """
        Run the report
        """
        original, synthetic = self._extract_report_data()
        (
            original,
            synthetic,
            float_columns,
            int_columns,
            categorical_columns,
            date_columns,
        ) = self.preprocess_data(original, synthetic)
        accuracy_test = AccuracyTest(
            original,
            synthetic,
            self.paths,
            self.table_name,
            self.config
        )
        accuracy_test.report(
            cont_columns=list(float_columns | int_columns),
            categorical_columns=list(categorical_columns),
            date_columns=list(date_columns),
        )


class SampleAccuracyReporter(Reporter):
    """
    Reporter for running a sample test
    """

    report_type = "sample"

    def _extract_report_data(self):
        original, schema = DataLoader(self.paths["source_path"]).load_data()
        sampled, schema = DataLoader(self.paths["input_data_path"], sensitive=True).load_data()
        return original, sampled

    def report(self):
        """
        Run the report
        """
        original, sampled = self._extract_report_data()
        (
            original,
            sampled,
            float_columns,
            int_columns,
            categorical_columns,
            date_columns,
        ) = self.preprocess_data(original, sampled)
        accuracy_test = SampleAccuracyTest(
            original,
            sampled,
            self.paths,
            self.table_name,
            self.config,
        )
        accuracy_test.report(
            cont_columns=list(float_columns | int_columns),
            categorical_columns=list(categorical_columns),
            date_columns=list(date_columns),
        )



from typing import Dict, List, Optional, Any, Callable, Literal
import os
import shutil
from itertools import product

import pandas as pd
from attrs import define, field
from copy import deepcopy
from loguru import logger
from slugify import slugify

from syngen.ml.data_loaders import MetadataLoader
from syngen.ml.strategies import TrainStrategy, InferStrategy
from syngen.ml.reporters import Report
from syngen.ml.config import Validator
from syngen.ml.mlflow_tracker import MlflowTrackerFactory
from syngen.ml.context.context import global_context
from syngen.ml.utils import ProgressBarHandler
from syngen.ml.mlflow_tracker import MlflowTracker
from syngen.ml.validation_schema import ReportTypes
from syngen.ml.processors import PreprocessHandler, PostprocessHandler


@define
class Worker:
    """
    Class for preparing training and infer settings, metadata for training and infer process
    """

    table_name: Optional[str] = field(kw_only=True)
    metadata_path: Optional[str] = field(kw_only=True)
    settings: Dict = field(kw_only=True)
    log_level: str = field(kw_only=True)
    type_of_process: Literal["train", "infer"] = field(kw_only=True)
    metadata: Optional[Dict] = None
    loader: Optional[Callable[[str], pd.DataFrame]] = None
    divided: List = field(default=list())
    initial_table_names: List = field(default=list())
    merged_metadata: Dict = field(default=dict())
    train_stages: List = ["PREPROCESS", "TRAIN", "POSTPROCESS"]
    infer_stages: List = ["INFER", "REPORT"]

    def __attrs_post_init__(self):
        os.makedirs("model_artifacts/metadata", exist_ok=True)
        self.metadata = self.__fetch_metadata()
        self._update_metadata()
        self._clean_up()
        self.__validate_metadata()
        self.initial_table_names = list(self.merged_metadata.keys())
        self._set_mlflow()

    def _clean_up(self):
        """
        Clean up the directories before the preprocessing data
        """
        for table in self.metadata.keys():
            if table != "global" and self.type_of_process == "train":
                self._remove_existed_artifacts(table)
                self._prepare_dirs(table)

    @staticmethod
    def _remove_existed_artifact(path_to_artifact: str):
        """
        Remove the existed artifact from the previous train process
        """
        if os.path.exists(path_to_artifact):
            shutil.rmtree(path_to_artifact)
            logger.info(f"The artifacts located in the path - '{path_to_artifact}' was removed")

    def _remove_existed_artifacts(self, table_name: str):
        """
        Remove existed artifacts from previous train process
        """
        resources_path = f"model_artifacts/resources/{slugify(table_name)}/"
        tmp_store_path = f"model_artifacts/tmp_store/{slugify(table_name)}/"
        self._remove_existed_artifact(resources_path)
        self._remove_existed_artifact(tmp_store_path)

    @staticmethod
    def _prepare_dirs(table_name: str):
        """
        Create main directories for saving original, synthetic data and model artifacts
        """
        resources_path = f"model_artifacts/resources/{slugify(table_name)}/"
        tmp_store_path = f"model_artifacts/tmp_store/{slugify(table_name)}/"
        state_path = (
            f"model_artifacts/resources/{slugify(table_name)}/vae/checkpoints"
        )
        flatten_config_path = "model_artifacts/tmp_store/flatten_configs/"
        os.makedirs(resources_path, exist_ok=True)
        os.makedirs(tmp_store_path, exist_ok=True)
        os.makedirs(state_path, exist_ok=True)
        os.makedirs(flatten_config_path, exist_ok=True)

    def __validate_metadata(self):
        """
        Validate the metadata, set the merged metadata
        """
        validator = Validator(
            metadata=self.metadata,
            metadata_path=self.metadata_path,
            type_of_process=self.type_of_process,
            validation_source=False if self.loader and self.type_of_process == "train" else True
        )
        validator.run()
        self.merged_metadata = validator.merged_metadata

    def __preprocess_data(self):
        """
        Preprocess the data before a training process
        """
        PreprocessHandler(
            metadata=self.metadata,
            metadata_path=self.metadata_path,
            table_name=self.table_name,
            loader=self.loader
        ).run()

    def __postprocess_data(self):
        """
        Postprocess the data after an inference process
        """
        PostprocessHandler(
            metadata=self.metadata,
            metadata_path=self.metadata_path,
            table_name=self.table_name
        ).run()

    def _set_mlflow(self):
        """
        Set the mlflow experiment name and the mlflow run
        """
        MlflowTrackerFactory.create_tracker(
            table_name=self.table_name,
            metadata_path=self.metadata_path
        )

    def _update_metadata_for_table(self):
        """
        Update the metadata for training or inference process if a metadata file wasn't provided
        """
        if self.type_of_process == "train":
            train_settings = self.metadata[self.table_name]["train_settings"]
            train_settings.update(self.settings)
        elif self.type_of_process == "infer":
            infer_settings = self.metadata[self.table_name]["infer_settings"]
            infer_settings.update(self.settings)

    @staticmethod
    def _update_table_settings(
        table_settings: Dict[str, Any], settings_to_update: Dict[str, Any]
    ) -> None:
        """
        Update the table settings with the provided settings that are not already defined
        """
        for setting, value in settings_to_update.items():
            if setting not in table_settings:
                table_settings[setting] = value

    def _update_metadata_for_tables(self):
        """
        Update the metadata for training or inference process if a metadata file was provided
        """
        global_train_settings = self.metadata.get("global", {}).get("train_settings", {})
        global_infer_settings = self.metadata.get("global", {}).get("infer_settings", {})

        for table, table_metadata in self.metadata.items():
            if table == "global":
                continue
            elif self.type_of_process == "train":
                settings_key = "train_settings"
                global_settings = global_train_settings
            elif self.type_of_process == "infer":
                settings_key = "infer_settings"
                global_settings = global_infer_settings
            else:
                continue

            table_settings = table_metadata[settings_key]

            self._update_table_settings(table_settings, global_settings)
            self._update_table_settings(table_settings, self.settings)

    def _update_metadata(self) -> None:
        if self.metadata_path:
            self._update_metadata_for_tables()
        if self.table_name:
            self._update_metadata_for_table()

    def __fetch_metadata(self) -> Dict:
        """
        Fetch the metadata for training or infer process
        """
        if self.metadata_path:
            metadata = MetadataLoader(self.metadata_path).load_data()
            return metadata
        if self.table_name:
            source = self.settings.get("source")
            metadata = {
                self.table_name: {
                    "train_settings": {
                        "source": source,
                    },
                    "infer_settings": {},
                    "keys": {},
                    "format": {}
                }
            }
            return metadata

    @staticmethod
    def _get_tables_without_keys(config_of_tables: Dict) -> List[str]:
        """
        Return the list of tables without the information about keys
        """
        tables_without_keys = []
        for table in config_of_tables.keys():
            if not config_of_tables[table].get("keys"):
                tables_without_keys.append(table)
        return tables_without_keys

    @staticmethod
    def _get_tables(config: Dict, key_type: str):
        """
        Return the list of related tables regarding the type of key -
        'primary key', 'foreign key', 'unique key'
        :param config: configuration of related tables declared in metadata.yaml file
        :param key_type:
        type of key either 'primary key' ('PK'), 'foreign key' ('FK'), 'unique key' ('UQ')
        """
        try:
            tables = [
                table_name
                for table_name, config in config.items()
                for key in config["keys"]
                if config["keys"][key]["type"] == key_type
            ]
            return list(dict.fromkeys(tables))
        except KeyError:
            raise KeyError(
                "The structure of metadata file seems to be invalid. "
                "Please check the presence of 'configuration', 'tables', "
                "'keys', 'type' fields in metadata file"
            )

    def _prepare_metadata_for_process(self, **kwargs):
        """
        Return the list of tables for training or infer process, the configuration of tables

        type_of_process can be "train", "infer" or "all" for the Enterprise version
        """
        config_of_tables = deepcopy(self.merged_metadata)
        if kwargs.get("type_of_process") in ("infer", "all"):
            config_of_tables = self._split_pk_fk_metadata(
                config_of_tables, list(config_of_tables.keys())
            )
        tables_without_keys = self._get_tables_without_keys(config_of_tables)
        pk_tables = self._get_tables(config_of_tables, "PK")
        uq_tables = self._get_tables(config_of_tables, "UQ")
        fk_tables = self._get_tables(config_of_tables, "FK")
        chain_of_tables = list(
            dict.fromkeys([*tables_without_keys, *pk_tables, *uq_tables, *fk_tables])
        )

        config_of_tables = {
            **self.metadata,
            **{k: v for k, v in config_of_tables.items() if k not in self.metadata},
        }
        chain_of_tables = [i for i in chain_of_tables for j in self.metadata if j in i]
        return chain_of_tables, config_of_tables

    @staticmethod
    def _split_keys(keys):
        """
        Split the keys into primary keys and unique keys, and foreign keys
        """
        pk_uq_keys = {k: v for k, v in keys.items() if v["type"] in ["PK", "UQ"]}
        fk_keys = {k: v for k, v in keys.items() if v["type"] == "FK"}
        return pk_uq_keys, fk_keys

    @staticmethod
    def _get_meta_copy(original_meta, new_keys):
        """
        Get updated metadata copy
        """
        meta_copy = original_meta.copy()
        meta_copy["keys"] = new_keys
        return meta_copy

    def _split_pk_fk_metadata(self, config, tables):
        """
        Split the metadata of tables into primary key and foreign key metadata
        """

        for table in tables:
            keys = config[table].get("keys", {})
            types_of_keys = {k_type["type"] for k_type in keys.values()}

            if "PK" in types_of_keys and "FK" in types_of_keys:
                self.divided.append(f"{table}_pk")
                self.divided.append(f"{table}_fk")

                pk_uq_keys, fk_keys = self._split_keys(keys)

                config[f"{table}_pk"] = self._get_meta_copy(config[table], pk_uq_keys)
                config[f"{table}_fk"] = self._get_meta_copy(config[table], fk_keys)

                config.pop(table)

        return config

    @staticmethod
    def _should_generate_data(
        config_of_tables: Dict,
        type_of_process: str,
        list_of_reports: List[str]
    ):
        """
        Determine whether the synthetic data should be generated
        in order to generate reports based on it
        """
        return any(
            [
                report in list_of_reports
                for config in config_of_tables.values()
                for report in config.get(f"{type_of_process}_settings", {}).get("reports", [])
            ]
        )

    def _collect_metrics_in_train(
        self,
        tables_for_training: List[str],
        tables_for_inference: List[str],
        generation_of_reports: bool
    ):
        """
        Collect the integral metrics for the training process
        """
        MlflowTracker().start_run(
            run_name="integral_metrics",
            tags={"process": "bottleneck"}
        )
        self._collect_integral_metrics(tables=tables_for_training, type_of_process="train")
        if generation_of_reports:
            self._collect_integral_metrics(tables=tables_for_inference, type_of_process="infer")
        MlflowTracker().end_run()

    def _train_table(self, table, metadata, delta):
        """"
        Train process for a single table
        """
        config_of_table = metadata[table]
        global_context(config_of_table.get("format", {}))
        train_settings = config_of_table["train_settings"]
        log_message = f"Training process of the table - '{table}' has started"
        logger.info(log_message)
        ProgressBarHandler().set_progress(delta=delta, message=log_message)

        TrainStrategy().run(
            metadata=metadata,
            metadata_path=self.metadata_path,
            source=train_settings.get("source"),
            epochs=train_settings["epochs"],
            drop_null=train_settings["drop_null"],
            row_limit=train_settings["row_limit"],
            table_name=table,
            reports=train_settings["reports"],
            batch_size=train_settings["batch_size"],
            loader=self.loader
        )
        self._write_success_file(slugify(table))
        self._save_metadata_file()
        ProgressBarHandler().set_progress(
            delta=delta,
            message=f"Training of the table - '{table}' was completed"
        )

    def __train_tables(
        self,
        tables_for_training: List,
        tables_for_inference: List,
        metadata_for_training: Dict,
        metadata_for_inference: Dict,
        generation_of_reports: bool
    ):
        """
        Run training process for the list of tables
        :param metadata_for_training:
        the tuple included the list of tables and metadata for training process
        :param metadata_for_inference:
        the tuple included the list of tables and metadata for inference process
        """
        delta = 0.49 / len(tables_for_training)

        self.__preprocess_data()

        for table in tables_for_training:
            self._train_table(table, metadata_for_training, delta)

        if generation_of_reports:
            self.__infer_tables(
                tables_for_inference,
                metadata_for_inference,
                delta,
                type_of_process="train"
            )

    def _get_surrogate_tables_mapping(self):
        """
        Get the mapping of surrogate tables, which end with "_pk" and "_fk",
        to the initial tables from which they were derived
        """
        return {
            table: [t for t in self.divided if t.startswith(table)]
            for table in self.initial_table_names
            if any(t.startswith(table) for t in self.divided)
        }

    def _find_parent_table(self, table):
        """
        Find the initial table
        from which the surrogate table was derived
        """
        return next(
            (
                parent
                for parent, children in self._get_surrogate_tables_mapping().items()
                if table in children
            ), None
        )

    def _infer_table(self, table, metadata, type_of_process, delta, is_nested=False):
        """
        Infer process for a single table
        """
        config_of_table = metadata[table]
        global_context(config_of_table.get("format", {}))
        log_message = f"Infer process of the table - '{table}' has started"
        logger.info(log_message)
        ProgressBarHandler().set_progress(delta=delta, message=log_message)
        both_keys = table in self.divided
        settings = config_of_table[f"{type_of_process}_settings"]

        MlflowTracker().start_run(
                run_name=f"{table}-INFER",
                tags={"table_name": table, "process": type_of_process},
                nested=is_nested,
        )
        InferStrategy().run(
            destination=settings.get("destination") if type_of_process == "infer" else None,
            metadata=metadata,
            metadata_path=self.metadata_path,
            size=settings.get("size") if type_of_process == "infer" else None,
            table_name=table,
            run_parallel=settings.get("run_parallel") if type_of_process == "infer" else False,
            batch_size=settings.get("batch_size") if type_of_process == "infer" else 1000,
            random_seed=settings.get("random_seed") if type_of_process == "infer" else 1,
            reports=settings["reports"],
            log_level=self.log_level,
            both_keys=both_keys,
            type_of_process=self.type_of_process,
            loader=self.loader
        )
        ProgressBarHandler().set_progress(
            delta=delta,
            message=f"Infer process of the table - '{table}' was completed"
        )
        MlflowTracker().end_run()

    def __infer_tables(
        self,
        tables: List,
        config_of_tables: Dict,
        delta: float,
        type_of_process: str
    ):
        """
        Run infer process for the list of tables
        :param tables: the list of tables for infer process
        :param config_of_tables: configuration of tables declared in metadata file
        """

        non_surrogate_tables = [table for table in tables if table not in self.divided]

        for table in non_surrogate_tables:
            self._infer_table(
                table=table,
                metadata=config_of_tables,
                type_of_process=type_of_process,
                delta=delta
            )

        tables_mapping = self._get_surrogate_tables_mapping()
        for table_root in tables_mapping.keys():
            MlflowTracker().start_run(
                run_name=f"{table_root}-INFER",
                tags={"table_name": table_root, "process": type_of_process}
            )
            for table in tables_mapping[table_root]:
                self._infer_table(
                    table=table,
                    metadata=config_of_tables,
                    type_of_process=type_of_process,
                    delta=delta,
                    is_nested=True
                )
            MlflowTracker().end_run()

    def _collect_integral_metrics(self, tables, type_of_process):
        """
        Collect the integral metrics depending on the type of process
        """
        stages = self.train_stages if type_of_process == "train" else self.infer_stages
        for table, stage in product(tables, stages):
            MlflowTracker().collect_metrics(table, stage)

    def _generate_reports(self):
        """
        Generate reports
        """
        Report().generate_report()
        Report().clear_report()

    @staticmethod
    def _write_success_file(table_name: str):
        """
        Write success message to the '.success' file
        """
        with open(f"model_artifacts/resources/{table_name}/message.success", "w") as f:
            f.write("SUCCESS")

    def _save_metadata_file(self):
        if self.metadata_path:
            os.makedirs("model_artifacts/metadata", exist_ok=True)
            metadata_file_name = os.path.basename(self.metadata_path)
            MetadataLoader(
                f"model_artifacts/metadata/{metadata_file_name}"
            ).save_data(self.metadata)

    def launch_train(self):
        """
        Launch training process either for a single table or for several tables
        """
        metadata_for_training = self._prepare_metadata_for_process(type_of_process="train")
        metadata_for_inference = self._prepare_metadata_for_process(type_of_process="infer")

        (
            tables_for_training,
            metadata_for_training,
        ) = metadata_for_training
        (
            tables_for_inference,
            metadata_for_inference,
        ) = metadata_for_inference

        generation_of_reports = self._should_generate_data(
            metadata_for_training,
            "train",
            ReportTypes().infer_report_types
        )

        self.__train_tables(
            tables_for_training,
            tables_for_inference,
            metadata_for_training,
            metadata_for_inference,
            generation_of_reports
        )

        self._generate_reports()
        self._collect_metrics_in_train(
            tables_for_training,
            tables_for_inference,
            generation_of_reports
        )

    def _collect_metrics_in_infer(self, tables):
        """
        Collect the integral metrics for the inference process
        """
        MlflowTracker().start_run(
            run_name="integral_metrics",
            tags={"process": "bottleneck"}
        )
        self._collect_integral_metrics(tables, type_of_process="infer")
        MlflowTracker().end_run()

    def launch_infer(self):
        """
        Launch infer process either for a single table or for several tables
        """
        tables, config_of_tables = self._prepare_metadata_for_process(type_of_process="infer")

        generation_of_reports = self._should_generate_data(
            config_of_tables,
            "infer",
            ReportTypes().infer_report_types
        )
        delta = 0.25 / len(tables) if generation_of_reports else 0.5 / len(tables)

        self.__infer_tables(tables, config_of_tables, delta, type_of_process="infer")
        self._generate_reports()
        self.__postprocess_data()
        self._collect_metrics_in_infer(tables)



import os
from typing import Optional, List
import traceback

import click
from loguru import logger

from syngen.ml.worker import Worker
from syngen.ml.utils import (
    setup_log_process,
    check_if_logs_available,
    validate_parameter_reports
)
from syngen.ml.validation_schema import ReportTypes


validate_reports = validate_parameter_reports(
    report_types=ReportTypes().infer_report_types,
    full_list=ReportTypes().full_list_of_infer_report_types
)


@click.command()
@click.option(
    "--metadata_path",
    type=str,
    default=None,
    help="Path to the metadata file"
)
@click.option(
    "--size",
    default=100,
    type=click.IntRange(1),
    help="Desired number of rows to generate. If absent, it's defaulted to 100",
)
@click.option(
    "--table_name",
    default=None,
    type=str,
    help="Name of the table, same as in training",
)
@click.option(
    "--run_parallel",
    default=False,
    type=click.BOOL,
    help="The flag which set whether to use multiprocessing (feasible for tables > 5000 rows)."
    "If absent, it's defaulted to False",
)
@click.option(
    "--batch_size",
    default=None,
    type=click.IntRange(1),
    help="If specified, the generation is split into batches. This can save the RAM",
)
@click.option(
    "--random_seed",
    default=None,
    type=click.IntRange(0),
    help="Set any int in case you want reproducible results. To reproduce generated data again, "
    "use the same int in this command.",
)
@click.option(
    "--reports",
    default=("none",),
    type=click.UNPROCESSED,
    multiple=True,
    callback=validate_reports,
    help="Controls the generation of quality reports. "
    "Might require significant time for big generated tables (>10000 rows). "
    "If set to 'accuracy', generates an accuracy report. "
    "If set to 'metrics_only', outputs the metrics information "
    "only to standard output without generation of a report. "
    "If set to 'all', generates an accuracy report. "
    "If it's absent or set to 'none', no reports are generated.",
)
@click.option(
    "--log_level",
    default="INFO",
    type=click.Choice(["TRACE", "DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]),
    help="Set the logging level which will be used in the process. "
         "If absent, it's defaulted to 'INFO'",
)
def launch_infer(
    metadata_path: Optional[str],
    size: Optional[int],
    table_name: Optional[str],
    run_parallel: bool,
    batch_size: Optional[int],
    reports: List[str],
    random_seed: Optional[int],
    log_level: str,
):
    """
    Launch the work of infer process
    Parameters
    ----------
    metadata_path
    size
    table_name
    run_parallel
    batch_size
    reports
    random_seed
    log_level
    -------

    """
    setup_log_process(
        type_of_process="infer",
        log_level=log_level,
        table_name=table_name,
        metadata_path=metadata_path
    )
    if not metadata_path and not table_name:
        raise AttributeError(
            "It seems that the information of 'metadata_path' or 'table_name' is absent. "
            "Please provide either the information of 'metadata_path' or the information "
            "of 'table_name'"
        )
    if metadata_path and table_name:
        logger.warning(
            "The information of 'metadata_path' was provided. "
            "In this case the information of 'table_name' will be ignored"
        )
        table_name = None
    logger.info(
        "The inference process will be executed according to the information mentioned "
        "in 'infer_settings' in the metadata file. If appropriate information is absent "
        "from the metadata file, then the values of parameters sent through CLI will be used. "
        "Otherwise, the values of parameters will be defaulted"
    )
    settings = {
        "size": size,
        "run_parallel": run_parallel,
        "batch_size": batch_size,
        "reports": reports,
        "random_seed": random_seed
    }
    worker = Worker(
        table_name=table_name,
        metadata_path=metadata_path,
        settings=settings,
        log_level=log_level,
        type_of_process="infer",
    )

    worker.launch_infer()


if __name__ == "__main__":
    try:
        launch_infer()
    except Exception as e:
        log_file = os.getenv("SUCCESS_LOG_FILE")
        if not os.path.exists(log_file):
            logger.error(
                f"Generation failed on running stage. "
                f"The traceback of the error - {traceback.format_exc()}"
            )
        raise e
    finally:
        check_if_logs_available()



import os
from pathlib import Path
from abc import ABC, abstractmethod
from typing import Optional, Dict, Tuple, List, Literal
import pickle as pkl
import csv
import inspect
from dataclasses import dataclass

import pandas as pd
import pandas.errors
import pandavro as pdx
import yaml
from yaml import SafeLoader
from yaml.scanner import ScannerError
from avro.errors import InvalidAvroBinaryEncoding
from loguru import logger
import fastavro
from cryptography.fernet import Fernet

from syngen.ml.validation_schema import SUPPORTED_EXCEL_EXTENSIONS
from syngen.ml.convertor import CSVConvertor, AvroConvertor
from syngen.ml.utils import trim_string
from syngen.ml.context import get_context, global_context
from syngen.ml.validation_schema import (
    ExcelFormatSettingsSchema,
    CSVFormatSettingsSchema,
    ReportTypes
)

DELIMITERS = {"\\t": "\t"}


class BaseDataLoader(ABC):
    """
    Abstract class for data loader
    """
    def __init__(self, path: str):
        if not path:
            raise ValueError("It seems that the information of source is absent")
        self.path = path

    @abstractmethod
    def load_data(self, *args, **kwargs):
        pass

    @abstractmethod
    def save_data(self, path: str, df: pd.DataFrame, **kwargs):
        pass


class DataLoader(BaseDataLoader):
    """
    Base class for loading and saving data
    """

    def __init__(self, path: str, sensitive: bool = False):
        super().__init__(path)
        fernet_key = os.getenv("FERNET_KEY")
        self.sensitive = (
            True if sensitive and fernet_key else False
        ) or self.path.endswith(".dat")
        self.file_loader = self._get_file_loader()
        self.has_existed_path = self.__check_if_path_exists()
        self.has_existed_destination = self.__check_if_path_exists(type_of_path="destination")

    @property
    def original_schema(self):
        return (
            self.file_loader.load_original_schema()
            if hasattr(self.file_loader, "load_original_schema")
            else None
        )

    def __check_if_path_exists(self, type_of_path="source"):
        if (type_of_path == "source" and os.path.exists(self.path)) or (
            type_of_path == "destination" and os.path.exists(os.path.dirname(self.path))
        ):
            return True
        return False

    def _get_file_loader(self):
        path = Path(self.path)
        if self.sensitive:
            return DataEncryptor(self.path)
        elif path.suffix == ".avro":
            return AvroLoader(self.path)
        elif path.suffix in [".csv", ".txt"]:
            return CSVLoader(self.path)
        elif path.suffix == ".tsv":
            return CSVLoader(self.path, sep="\t")
        elif path.suffix == ".psv":
            return CSVLoader(self.path, sep="|")
        elif path.suffix == ".pkl":
            return BinaryLoader(self.path)
        elif path.suffix in SUPPORTED_EXCEL_EXTENSIONS:
            return ExcelLoader(self.path)
        else:
            raise NotImplementedError(f"File format not supported for extension {path.suffix}")

    def load_data(self, **kwargs) -> Tuple[pd.DataFrame, Dict]:
        try:
            df, schema = self.file_loader.load_data(**kwargs)
            return df, schema
        except UnicodeDecodeError as error:
            message = (
                f"It seems that the content of the data in the path - '{self.path}' "
                f"doesn't have the encoding UTF-8. The details of the error - {error}.\n"
                f"Please, use the data in UTF-8 encoding"
            )
            logger.error(message, error)
            raise error
        except pandas.errors.EmptyDataError as error:
            message = (
                f"The empty file was provided. Unable to load data from the path - '{self.path}'. "
                f"The details of the error - {error}"
            )
            logger.error(message, error)
            raise error

    def save_data(self, data, **kwargs):
        if data is not None:
            self.file_loader.save_data(data, **kwargs)

    def get_columns(self, **kwargs) -> List[str]:
        return self.file_loader.get_columns(**kwargs)


class CSVLoader(BaseDataLoader):
    """
    Class for loading and saving data in '.csv' format.
    """

    def __init__(self, path: str, **kwargs):
        super().__init__(path)
        self.format = get_context().get_config()
        self.format.update(kwargs)
        self.format = {
            k: v
            for k, v in self.format.items()
            if k in CSVFormatSettingsSchema._declared_fields.keys()
        }

    @staticmethod
    def _get_quoting(quoting: Optional[str]) -> int:
        quoting_map = {
            "minimal": csv.QUOTE_MINIMAL,
            "all": csv.QUOTE_ALL,
            "non-numeric": csv.QUOTE_NONNUMERIC,
            "none": csv.QUOTE_NONE,
        }
        if isinstance(quoting, int):
            return quoting
        else:
            return (
                quoting_map.get(quoting.lower(), csv.QUOTE_MINIMAL)
                if quoting
                else csv.QUOTE_MINIMAL
            )

    @staticmethod
    def _get_csv_params(**kwargs):
        params = {}
        format_params = kwargs.get("format")

        if format_params:
            params.update(format_params)
            quoting = format_params.get("quoting", None)
            params["quoting"] = CSVLoader._get_quoting(quoting)

        return params

    def _fetch_data(self, **params):
        return pd.read_csv(self.path, **params).apply(trim_string, axis=0)

    def _load_data(self, **kwargs) -> Tuple[pd.DataFrame, Dict]:
        params = CSVLoader._get_csv_params(**kwargs)
        try:
            df = self._fetch_data(**params)
            if all([isinstance(column, int) for column in df.columns]):
                df.rename(
                    columns={
                        k: f"column_{v}" for k, v in zip(df.columns, list(range(len(df.columns))))
                    },
                    inplace=True,
                )
            sep = params.get("sep", ",")
            if len(sep) > 1:
                params["sep"] = ","
            params["skiprows"] = None
            global_context(params)
        except FileNotFoundError as error:
            message = (
                f"It seems that the path to the table isn't valid.\n"
                f"The details of the error - {error}.\n"
                f"Please, check the path to the table"
            )
            logger.error(message)
            raise FileNotFoundError(message)

        return df, CSVConvertor(df).schema

    def load_data(self, **kwargs):
        return self._load_data(format=self.format, **kwargs)

    def __get_columns(self, **kwargs):
        head_df = pd.read_csv(self.path, **kwargs, nrows=0)
        return list(head_df.columns)

    def get_columns(self, **kwargs) -> List[str]:
        return self._get_columns(**kwargs)

    def _get_columns(self, **kwargs) -> List[str]:
        """
        Get the column names of the table located in the path
        """
        try:
            return self.__get_columns(**kwargs)
        except pd.errors.EmptyDataError as error:
            logger.error(
                f"The empty file was provided. Unable to train this table located "
                f"in the path - '{self.path}'. The details of the error - {error}"
            )
            raise error

    def _write_data(self, df, **kwargs):
        """
        Save the dataframe in '.csv' format
        """
        df.to_csv(self.path, **kwargs, index=False)

    def _save_data(self, df: pd.DataFrame, **kwargs):
        """
        Save the provided DataFrame to a CSV file.
        :param path: The file path to save the DataFrame.
        :param df: The DataFrame to be saved.
        :param kwargs: Additional keyword arguments to be passed to the to_csv method.
        """
        format_params = CSVLoader._get_csv_params(**kwargs)
        if df is not None:
            # Extract valid parameters
            valid_parameters = inspect.signature(pd.DataFrame.to_csv).parameters

            # Filter out any keyword arguments that are not valid parameters
            filtered_kwargs = {k: v for k, v in format_params.items() if k in valid_parameters}
            for k, v in filtered_kwargs.items():
                if isinstance(v, str) and v in DELIMITERS.keys():
                    filtered_kwargs[k] = v.replace(v, DELIMITERS[v])

            if "header" in filtered_kwargs and filtered_kwargs.get("header", None) is None:
                filtered_kwargs["header"] = False
            else:
                filtered_kwargs["header"] = True

            if "sep" in filtered_kwargs and len(filtered_kwargs.get("sep", None)) > 1:
                filtered_kwargs["sep"] = ","
                logger.warning(
                    "As the length of the value of the parameter 'separator' is more than "
                    "1 character, the 'separator' will be set to ',' in accordance with "
                    "the standard 'RFC 4180'"
                )
            if (
                "na_values" in format_params
                and format_params.get("na_values", [])
                and df.isnull().values.any()
            ):
                filtered_kwargs["na_rep"] = format_params["na_values"][0]
                logger.warning(
                    "Since the 'na_values' parameter in the 'format' sections is not empty, "
                    "the missing values will be filled with "
                    "the first value from the 'na_values' parameter"
                )

            self._write_data(df, **filtered_kwargs)

    def save_data(self, df: pd.DataFrame, **kwargs):
        self._save_data(df, **kwargs)


class AvroLoader(BaseDataLoader):
    """
    Class for loading and saving data in '.avro' format
    """

    def _load_data(self) -> pd.DataFrame:
        """
        Load data in '.avro' format
        """
        with open(self.path, "rb") as f:
            return pdx.from_avro(f)

    @staticmethod
    def _get_preprocessed_schema(schema: Optional[Dict]) -> Optional[Dict]:
        """
        Get the preprocessed schema
        """
        if schema is not None:
            return {
                field["name"]: field["type"]
                for field
                in schema.get("fields", {})
            }
        return schema

    def load_data(self, **kwargs) -> Tuple[pd.DataFrame, Dict]:
        """
        Load data in '.avro' format
        """
        try:
            df = self._load_data()
            schema = self.load_schema()
            return self._get_schema_and_df(schema, df)
        except FileNotFoundError as error:
            message = (
                f"It seems that the path to the table isn't valid.\n"
                f"The details of the error - {error}.\n"
                f"Please, check the path to the table"
            )
            logger.error(message)
            raise FileNotFoundError(message)

    def _save_data(self, df: pd.DataFrame, schema: Optional[Dict]):
        """
        Save data in Avro Format
        """
        if df is not None:
            pdx.to_avro(self.path, df, schema)

    def save_data(self, df: pd.DataFrame, schema: Optional[Dict] = None, **kwargs):
        if schema is not None:
            logger.trace(f"The data will be saved with the schema: {schema}")

        preprocessed_schema = self._get_preprocessed_schema(schema)
        df = AvroConvertor(preprocessed_schema, df).preprocessed_df
        self._save_data(df, schema)

    def __load_original_schema(self):
        with open(self.path, "rb") as f:
            return self._load_original_schema(f)

    @staticmethod
    def _load_original_schema(f):
        """
        Load schema of the metadata of the table in '.avro' format
        """
        reader = fastavro.reader(f)
        return reader.writer_schema

    def load_original_schema(self) -> Dict:
        return self.__load_original_schema()

    def load_schema(self) -> Dict[str, str]:
        """
        Load the custom schema of the metadata of the table in '.avro' format
        used by the algorithm
        """
        original_schema = self.load_original_schema()
        return self._get_preprocessed_schema(original_schema)

    @staticmethod
    def _get_schema_and_df(
        schema: Dict[str, str], df: pd.DataFrame
    ) -> Tuple[pd.DataFrame, Dict[str, str]]:
        """
        Preprocess schema and dataframe
        """
        convertor = AvroConvertor(schema, df)
        schema, preprocessed_df = convertor.converted_schema, convertor.preprocessed_df
        return preprocessed_df, schema

    def _get_columns(self) -> List[str]:
        """
        Get the column names of the table located in the path
        """
        schema = self.load_schema()
        return list(schema.keys())

    def get_columns(self, **kwargs) -> List[str]:
        try:
            return self._get_columns()
        except InvalidAvroBinaryEncoding as error:
            logger.error(
                f"The empty file was provided. Unable to train this table "
                f"located in the path - '{self.path}'. "
                f"The details of the error - {error}"
            )
            raise error


class MetadataLoader(BaseDataLoader):
    """
    Metadata class for loading and saving metadata in YAML format
    """

    def __init__(self, path: str):
        super().__init__(path)
        self.metadata_loader = self.get_metadata_loader()

    def get_metadata_loader(self):
        if self.path is not None:
            path = Path(self.path)
            if path.suffix in [".yaml", ".yml"]:
                return YAMLLoader(self.path)
            else:
                raise NotImplementedError("The format of metadata isn't supported")

    def load_data(self) -> Dict:
        return self.metadata_loader.load_data()

    def save_data(self, metadata: Dict, **kwargs):
        self.metadata_loader.save_data(metadata, **kwargs)


@dataclass
class YAMLLoader(BaseDataLoader):
    """
    Class for loading and saving data in YAML format
    """
    metadata_sections = ["train_settings", "infer_settings", "format", "keys"]
    train_reports = ReportTypes().full_list_of_train_report_types
    infer_reports = ReportTypes().full_list_of_infer_report_types

    def __init__(self, path: str):
        super().__init__(path)

    def _normalize_reports(self, settings: dict, type_of_process: Literal["train", "infer"]):
        """
        Cast the value of the parameter 'reports' to the list
        """
        reports = settings.get(f"{type_of_process}_settings", {}).get("reports", [])
        if isinstance(reports, str):
            if reports not in ["none", "all"]:
                settings[f"{type_of_process}_settings"]["reports"] = [reports]
            elif reports == "none":
                settings[f"{type_of_process}_settings"]["reports"] = []
            elif reports == "all" and type_of_process == "train":
                settings[f"{type_of_process}_settings"]["reports"] = self.train_reports
            elif reports == "all" and type_of_process == "infer":
                settings[f"{type_of_process}_settings"]["reports"] = self.infer_reports

    def _normalize_parameter_reports(self, metadata: dict) -> dict:
        for table, settings in metadata.items():
            self._normalize_reports(settings, "train")
            self._normalize_reports(settings, "infer")
        return metadata

    def _load_data(self, metadata_file) -> Dict:
        try:
            metadata = yaml.load(metadata_file, Loader=SafeLoader)
            metadata = self.replace_none_values_of_metadata_settings(metadata)
            metadata = self._normalize_parameter_reports(metadata)
            return metadata
        except ScannerError as error:
            message = (
                f"It seems that the metadata file in YAML format isn't valid.\n"
                f"The details of the error - {str(error)}.\n"
                f"Please check the metadata file in YAML format."
            )
            logger.error(error)
            raise ValueError(message)

    def load_data(self) -> Dict:
        with open(self.path, "r", encoding="utf-8") as f:
            return self._load_data(f)

    def replace_none_values_of_metadata_settings(self, metadata: dict):
        """
        Replace None values for parameters in the metadata
        """
        metadata["global"] = dict() if metadata.get("global") is None else metadata.get("global")
        if metadata["global"]:
            for settings in metadata["global"].keys():
                metadata["global"][settings] = (
                    dict()
                    if metadata["global"].get(settings) is None
                    else metadata["global"].get(settings)
                )
        for key in metadata.keys():
            if key == "global":
                continue
            for parameter in self.metadata_sections:
                if metadata.get(key).get(parameter) is None:
                    metadata[key][parameter] = {}
        return metadata

    def save_data(self, metadata: Dict, **kwargs):
        with open(self.path, "w") as f:
            self._save_data(metadata, f)

    @staticmethod
    def _save_data(metadata: Dict, f):
        yaml.dump(metadata, f)


class BinaryLoader(BaseDataLoader):
    """
    Class for loading and saving data using byte stream
    """

    def _load_data(self) -> pd.DataFrame:
        """
        Load data in Binary format
        """
        with open(self.path, "rb") as f:
            return pkl.load(f)

    def get_columns(self) -> List[str]:
        data, schema = self.load_data()
        return data.columns.tolist()

    def load_data(self) -> Tuple[pd.DataFrame, None]:
        return self._load_data(), None

    def _save_data(self, data):
        with open(self.path, "wb") as f:
            pkl.dump(data, f)

    def save_data(self, data, **kwargs):
        """
        Save data in Binary format
        """
        self._save_data(data)


class ExcelLoader(BaseDataLoader):
    """
    Class for loading and saving data in Excel format
    """

    def __init__(self, path: str):
        super().__init__(path)
        self.format = get_context().get_config()
        self.sheet_name = self.format.get("sheet_name", 0)
        self.format = {
            k: v
            for k, v in self.format.items()
            if k in ExcelFormatSettingsSchema._declared_fields.keys()
        }

    def _fetch_data(self) -> pd.DataFrame:
        return pd.read_excel(self.path, sheet_name=self.sheet_name)

    def _load_data(self) -> Tuple[pd.DataFrame, Dict]:
        """
        Load data in Excel format
        """
        try:
            df = self._fetch_data()
            if isinstance(self.sheet_name, list) or self.sheet_name is None:
                dfs = [df for sheet_name, df in df.items()]
                df = pd.concat(dfs, ignore_index=True)
            global_context({})
            return df, CSVConvertor(df).schema
        except FileNotFoundError as error:
            message = (
                f"It seems that the path to the table isn't valid.\n"
                f"The details of the error - {error}.\n"
                f"Please, check the path to the table"
            )
            logger.error(message, error)
            raise error

    def load_data(self) -> Tuple[pd.DataFrame, Dict]:
        return self._load_data()

    def save_data(self, df: pd.DataFrame, **kwargs):
        self._save_data(df)

    def _save_data(self, df: pd.DataFrame):
        """
        Save provided data frame in Excel format
        """
        if df is not None:
            df.to_excel(self.path, index=False, engine="openpyxl")

    def _get_columns(self, **kwargs) -> List[str]:
        """
        Get the column names of the table located in the path
        """
        head_df = pd.read_excel(self.path, **kwargs, nrows=0)
        return list(head_df.columns)

    def get_columns(self, **kwargs) -> List[str]:
        return self._get_columns(**kwargs)


class DataEncryptor(BaseDataLoader):
    """
    A class to handle encryption and decryption of data using a Fernet key
    """

    def __init__(self, path: str):
        """
        Initialize the DataEncryptor with a Fernet key.
        """
        super().__init__(path)
        key = os.getenv("FERNET_KEY")
        self._validate_fernet_key(key)
        self.fernet = Fernet(key)

    @staticmethod
    def _validate_fernet_key(key: str):
        """
        Validate the provided Fernet key.
        A valid Fernet key is a 44-character URL-safe base64-encoded string.
        """
        if key is None or not key.strip():
            raise ValueError("It seems that the Fernet key is absent")

        error_message = "It seems that the provided Fernet key is invalid"
        try:
            Fernet(key.encode())

        except ValueError as e:
            logger.error(f"{error_message}. {str(e)}")
            raise e

    def get_columns(self) -> List[str]:
        """
        Get the column names of the table
        """
        data, schema = self.load_data()
        return data.columns.tolist()

    def save_data(self, df: pd.DataFrame):
        """
        Save the encrypted dataframe to the disk
        """
        try:
            serialized_df: bytes = pkl.dumps(df, protocol=pkl.HIGHEST_PROTOCOL)
            encrypted_data = self.fernet.encrypt(serialized_df)

            # Use atomic write operation for better safety
            temp_path = f"{self.path}.tmp"
            with open(temp_path, "wb") as encrypted_file:
                encrypted_file.write(encrypted_data)
            os.replace(temp_path, self.path)

            logger.info(
                f"Data is successfully encrypted and saved to '{self.path}'."
            )
        except Exception as e:
            if os.path.exists(temp_path):
                os.remove(temp_path)
            logger.error(f"Encryption failed: {str(e)}")
            raise e

    def load_data(self, *args, **kwargs) -> Tuple[pd.DataFrame, Dict]:
        """
        Load the decrypted data from the disk
        """
        try:
            with open(self.path, "rb") as encrypted_file:
                encrypted_data = encrypted_file.read()

            decrypted_data = self.fernet.decrypt(encrypted_data)
            df_decrypted = pkl.loads(decrypted_data)

            logger.info(
                f"Data stored at the path - '{self.path}' "
                f"has been successfully decrypted and loaded."
            )
            return df_decrypted, {"fields": {}, "format": "CSV"}
        except Exception as e:
            logger.error(
                f"It seems that the decryption process of the data "
                f"stored at the path - '{self.path}' failed due to the following reasons - "
                f"the provided Fernet key is invalid or the encrypted data is corrupted. {str(e)}"
            )
            raise e



import os
from datetime import datetime
from typing import Tuple, List, Optional, Dict
from abc import ABC, abstractmethod
from pathlib import Path
from dataclasses import dataclass, field

import warnings
import pickle
import tensorflow as tf
from tensorflow.python.data.experimental import AutoShardPolicy
from tensorflow.keras.models import Model
import matplotlib.pyplot as plt
import time
import tqdm
import pandas as pd
import numpy as np
from loguru import logger
from slugify import slugify

from syngen.ml.vae.models.model import CVAE
from syngen.ml.vae.models import Dataset
from syngen.ml.mlflow_tracker import MlflowTracker
from syngen.ml.utils import (
    generate_uuid,
    fetch_config,
    check_if_features_assigned,
    ProgressBarHandler
)
from syngen.ml.data_loaders import DataLoader

warnings.filterwarnings("ignore")

BATCH_SIZE_DEFAULT = 32


class BaseWrapper(ABC):
    """
    Abstract class for wrappers
    """

    @abstractmethod
    def fit_on_df(self, df: pd.DataFrame, epochs: int, columns_subset: List[str] = None):
        pass

    @abstractmethod
    def predict_sampled_df(self, n: int) -> pd.DataFrame:
        pass

    @abstractmethod
    def save_state(self, path: str):
        pass


@dataclass
class VAEWrapper(BaseWrapper):
    df: pd.DataFrame
    schema: Optional[Dict]
    metadata: Dict
    table_name: str
    paths: Dict
    process: str
    main_process: str
    batch_size: int
    log_level: str
    losses_info: pd.DataFrame = field(init=True, default_factory=pd.DataFrame)
    dataset: Dataset = field(init=False)
    vae: CVAE = field(init=False, default=None)
    model: Model = field(init=False, default=None)
    num_batches: int = field(init=False)
    feature_types: Dict = field(init=False, default_factory=dict)

    def __post_init__(self):
        if self.process == "train":
            self._prepare_dir()
            self.dataset = Dataset(
                df=self.df,
                schema=self.schema,
                metadata=self.metadata,
                table_name=self.table_name,
                main_process=self.main_process,
                paths=self.paths,
            )
            self.dataset.launch_detection()
            self.df = self.dataset.pipeline()
            self._save_dataset()
        elif self.process == "infer":
            self.dataset = fetch_config(self.paths["dataset_pickle_path"])
            self._update_dataset()
            self._save_dataset()

    def _update_dataset(self):
        """
        Update dataset object related to the current process
        """
        self.dataset.paths = self.paths
        self.dataset.metadata = self.metadata
        self.dataset.main_process = self.main_process

    def _save_dataset(self):
        """
        Save dataset object on the disk
        """
        with open(self.paths["dataset_pickle_path"], "wb") as f:
            f.write(pickle.dumps(self.dataset))

    @staticmethod
    def _prepare_dir():
        os.makedirs("model_artifacts/tmp_store/losses", exist_ok=True)

    def _restore_zero_values(self, df):
        for column in self.dataset.zero_num_column_names:
            if column.endswith("_zero"):
                # remove _zero to get original column name
                num_column_name = column[:-5]
                num_column = df[num_column_name].copy()
                zero_column_mask = df[column].astype("float") >= 0.5
                num_column = num_column.where(zero_column_mask, 0)
                num_zero_values = (num_column == 0).sum()
                df[num_column_name] = num_column
                df = df.drop(column, axis=1)
                logger.info(
                    f"Column {num_column_name} has {num_zero_values} "
                    f"({round(num_zero_values * 100 / len(num_column), 2)}%) "
                    f"zero values generated"
                )
        return df

    def _restore_nan_values(self, df):
        for column in self.dataset.null_num_column_names:
            if column.endswith("_null"):
                # remove _null to get original column name
                num_column_name = column[:-5]
                num_column = df[num_column_name].copy()
                not_null_column_mask = df[column].astype("float64") <= 0.5
                num_column = num_column.where(not_null_column_mask, np.nan)
                df[num_column_name] = num_column
                df = df.drop(column, axis=1)
                num_nan_values = num_column.isna().sum()
                logger.info(
                    f"Column {num_column_name} has {num_nan_values} "
                    f"({round(num_nan_values * 100 / len(num_column), 2)}%) "
                    f"empty values generated"
                )
        return df

    def _restore_nan_labels(self, df):
        for column_name, nan_label in self.dataset.nan_labels_dict.items():
            if nan_label is None:
                nan_label = np.nan
            if column_name in df.select_dtypes(int).columns:
                df[column_name] = df[column_name].astype("object")
            df[column_name] = df[column_name].fillna(nan_label)
        return df

    def fit_on_df(
        self,
        epochs: int,
        columns_subset: List[str] = None,  # TODO columns_subset does not work
    ):
        if not check_if_features_assigned(self.paths["dataset_pickle_path"]):
            return

        if columns_subset is None:
            columns_subset = self.df.columns
        else:
            # if a column is in columns_subset, its null column should also be added if present
            columns_subset += [
                col
                for col in self.df.columns
                if col.endswith(("_null", "_zero")) and (col[:-5] in columns_subset)
            ]

        df = self.df.loc[:, list(set(columns_subset))]

        train_dataset = self._create_batched_dataset(df)
        self.num_batches = len(train_dataset)
        self.model = self.vae.model
        self.feature_types = self.vae.feature_types

        self.optimizer = self.__create_optimizer()
        self.loss_metric = self._create_loss()

        # Start of the run of training process
        MlflowTracker().start_run(
            run_name=f"{self.table_name}-TRAIN",
            tags={"table_name": self.table_name, "process": "train"},
        )
        config = fetch_config(self.paths["train_config_pickle_path"])
        MlflowTracker().log_params(config.to_dict())

        self._train(train_dataset, epochs)

        MlflowTracker().end_run()

        MlflowTracker().start_run(
            run_name=f"{self.table_name}-POSTPROCESS",
            tags={"table_name": self.table_name, "process": "postprocess"},
        )
        self.fit_sampler(df)

    def _calculate_loss_by_type(
        self,
        feature_losses: Dict,
        feature_type: str
    ) -> float:
        """
        Group features and calculate the loss by the type
        """
        return sum(
            loss for name, loss in feature_losses.items()
            if self.feature_types[name] == feature_type
        )

    def _get_grouped_losses(self, feature_losses, epoch) -> Dict:
        """
        Get the mean numerical, categorical, and text losses for every epoch
        """
        num_loss = self._calculate_loss_by_type(
            feature_losses,
            feature_type="numeric"
        )

        categorical_loss = self._calculate_loss_by_type(
            feature_losses,
            feature_type="categorical"
        )

        text_loss = self._calculate_loss_by_type(
            feature_losses,
            feature_type="text"
        )
        logger.trace(
            f"The 'numeric_loss' - {num_loss}, "
            f"the 'categorical_loss' - {categorical_loss}, "
            f"the 'text_loss' - {text_loss} in the {epoch} epoch"
        )
        MlflowTracker().log_metric(
            "numeric_loss", num_loss, step=epoch
        )
        MlflowTracker().log_metric(
            "categorical_loss", categorical_loss, step=epoch
        )
        MlflowTracker().log_metric(
            "text_loss", text_loss, step=epoch
        )
        return {
            "numeric_loss": num_loss,
            "categorical_loss": categorical_loss,
            "text_loss": text_loss
        }

    def _get_mean_feature_losses(self, total_feature_losses: Dict):
        """
        Get the mean loss of every feature
        """
        return {
            name: np.mean(loss / self.num_batches)
            for name, loss in total_feature_losses.items()
        }

    def _get_ending(self, feature_name):
        """
        Get the appropriate ending for the name of the loss of the certain feature
        """
        endings = {
            "categorical": "cat",
            "numeric": "num",
            "text": "text"
        }
        feature_type = self.feature_types.get(feature_name)
        return endings.get(feature_type)

    def _monitor_feature_losses(self, mean_feature_losses, epoch):
        """
        Monitor the mean value of the loss of every feature for every epoch
        """
        for name, loss in mean_feature_losses.items():
            ending = self._get_ending(feature_name=name)
            MlflowTracker().log_metric(
                f"{slugify(name)}_loss_{ending}", loss, step=epoch
            )

    def _fetch_feature_losses_info(
            self,
            feature_losses: Dict,
            epoch: int
    ) -> pd.DataFrame:
        """
        Fetch the information related to the loss for every feature in a certain epoch
        """
        timestamp = slugify(datetime.now().strftime("%Y-%m-%d %H:%M:%S"))
        table_name = f"{self.table_name}_{timestamp}"

        rows = []
        for name, loss in feature_losses.items():
            ending = self._get_ending(feature_name=name)
            row = {
                "table_name": table_name,
                "epoch": epoch,
                "column_name": name if ending else "",
                "column_type": self.feature_types.get(name, "general"),
                "loss_name": f"{name}_loss_{ending}" if ending else name,
                "value": loss,
            }
            rows.append(row)

        return pd.DataFrame(rows)

    @staticmethod
    def _accumulate_feature_losses(feature_losses: Dict, total_feature_losses: Dict):
        """
        Accumulate the loss for every feature
        """
        for key, value in feature_losses.items():
            if key in total_feature_losses:
                total_feature_losses[key] += value
            else:
                total_feature_losses[key] = value
        return total_feature_losses

    def _update_losses_info(self, feature_losses, epoch):
        """
        Add the information about losses of all features fetched during the certain epoch
        """
        data = self._fetch_feature_losses_info(feature_losses, epoch)
        self.losses_info = pd.concat([self.losses_info, data])

    def __save_losses(self):
        """
        Save the information about losses of every feature in every epoch
        """
        DataLoader(self.paths["losses_path"]).save_data(self.losses_info)

    def _gather_losses_info(self, total_feature_losses, mean_loss, mean_kl_loss, epoch):
        """
        Gather the information of losses related to every feature,
        numeric loss, categorical loss and text loss
        """
        mean_feature_losses = self._get_mean_feature_losses(total_feature_losses)
        self._update_losses_info(mean_feature_losses, epoch)
        self._monitor_feature_losses(
            mean_feature_losses,
            epoch
        )
        losses = self._get_grouped_losses(mean_feature_losses, epoch)
        losses.update({"kl_loss": mean_kl_loss, "total_loss": mean_loss})
        self._update_losses_info(losses, epoch)

    def _log_losses_info_to_mlflow(self):
        """
        Log 'losses.csv' to mlflow
        """
        path = self.paths["losses_path"]
        try:
            MlflowTracker().log_artifact(path)
        except Exception as error:
            logger.warning(
                f"Logging the report to mlflow has failed due to a permission error. "
                f"Error details: {error}.\n"
                f"The report will be saved locally in '{path}'"
            )
            pass

    def _train(self, dataset, epochs: int):
        step = self._train_step

        loss_grows_num_epochs = 0
        prev_total_loss = float("inf")
        es_min_delta = 0.005
        es_patience = 10
        pth = Path(self.paths["state_path"])
        # loss that corresponds to the best saved weights
        saved_weights_loss = float("inf")

        delta = ProgressBarHandler().delta / (epochs * 2)
        for epoch in range(epochs):
            log_message = (
                f"Training process of the table - '{self.table_name}' "
                f"on the epoch: {epoch}"
            )
            ProgressBarHandler().set_progress(
                progress=ProgressBarHandler().progress + delta,
                message=log_message
            )
            total_loss = 0.0
            total_feature_losses = dict()
            total_kl_loss = 0.0
            t1 = time.time()

            # Iterate over the batches of the dataset.
            for i, x_batch_train in tqdm.tqdm(iterable=enumerate(dataset)):
                loss, kl_loss, feature_losses = step(x_batch_train)
                total_loss += loss
                total_kl_loss += kl_loss
                total_feature_losses = self._accumulate_feature_losses(
                    feature_losses,
                    total_feature_losses
                )

            mean_loss = np.mean(total_loss / self.num_batches)
            mean_kl_loss = np.mean(total_kl_loss / self.num_batches)

            if mean_loss >= prev_total_loss - es_min_delta:
                loss_grows_num_epochs += 1
            else:
                self.model.save_weights(str(pth / "vae_best_weights_tmp.ckpt"))
                loss_grows_num_epochs = 0
                # loss that corresponds to the best saved weights
                saved_weights_loss = mean_loss

            log_message = (
                f"epoch: {epoch}, total loss: {mean_loss}, time: {(time.time() - t1):.4f} sec"
            )
            logger.info(log_message)

            ProgressBarHandler().set_progress(
                progress=ProgressBarHandler().progress + delta,
                message=log_message
            )

            self._gather_losses_info(total_feature_losses, mean_loss, mean_kl_loss, epoch)
            logger.trace(f"The 'kl_loss' - {mean_kl_loss} in {epoch} epoch")

            MlflowTracker().log_metric("loss", mean_loss, step=epoch)
            MlflowTracker().log_metric("saved_weights_loss", saved_weights_loss, step=epoch)
            MlflowTracker().log_metric("kl_loss", mean_kl_loss, step=epoch)

            prev_total_loss = mean_loss

            if loss_grows_num_epochs == es_patience:
                self.model.load_weights(str(pth / "vae_best_weights_tmp.ckpt"))
                logger.info(
                    f"The loss does not become lower for "
                    f"{loss_grows_num_epochs} epochs in a row. "
                    f"Stopping the training."
                )
                break
            epoch += 1
        self.__save_losses()
        self._log_losses_info_to_mlflow()

    @staticmethod
    def _create_optimizer(learning_rate):
        import platform
        if platform.processor() == 'arm':
            logger.info('Mac ARM processor is detected. Legacy Adam optimizer has been created.')
            return tf.keras.optimizers.legacy.Adam(learning_rate=learning_rate)
        else:
            return tf.keras.optimizers.Adam(learning_rate=learning_rate)

    def __create_optimizer(self):
        learning_rate = 1e-04 * np.sqrt(self.batch_size / BATCH_SIZE_DEFAULT)
        return self._create_optimizer(learning_rate)

    @staticmethod
    def _create_loss():
        return tf.keras.metrics.Mean()

    def _create_batched_dataset(self, df: pd.DataFrame):
        """
        Define batched dataset for training vae
        """
        transformed_data = self.dataset.transform(df)

        feature_datasets = []
        options = tf.data.Options()
        options.experimental_distribute.auto_shard_policy = AutoShardPolicy.DATA
        for inp in transformed_data:
            dataset = tf.data.Dataset.from_tensor_slices(inp).with_options(options)
            feature_datasets.append(dataset)

        dataset = tf.data.Dataset.zip(tuple(feature_datasets)).with_options(options)
        return dataset.batch(self.batch_size, drop_remainder=True)

    def _train_step(self, batch: Tuple[tf.Tensor]):
        with tf.GradientTape() as tape:
            self.model(batch)

            # Compute reconstruction loss
            loss = sum(self.model.losses)
            order_of_features = list(self.vae.feature_losses.keys())
            kl_loss = self.model.losses[-1].numpy()
            feature_losses = {
                name: loss.numpy()
                for name, loss in
                zip(order_of_features, self.model.losses[:-1])
            }

        self.optimizer.minimize(
            loss=loss,
            var_list=self.model.trainable_weights,
            tape=tape
        )
        self.loss_metric(loss)
        return loss, kl_loss, feature_losses

    @staticmethod
    def display_losses(feature_losses: Dict):
        for name, l in feature_losses.items():
            plt.plot(l, label=name)

        plt.legend()
        plt.ylim(0, 10)
        return plt.show()

    def fit_sampler(self, df: pd.DataFrame):
        self.vae.fit_sampler(df)

    def predict_sampled_df(self, n: int) -> pd.DataFrame:
        sampled_df = self.vae.sample(n)

        # uuid columns are generated here to restore nan values
        uuid_columns = self.dataset.uuid_columns
        if uuid_columns:
            sampled_df = generate_uuid(n, self.dataset,
                                       uuid_columns, sampled_df)

        sampled_df = self._restore_nan_values(sampled_df)
        sampled_df = self._restore_zero_values(sampled_df)
        sampled_df = self._restore_nan_labels(sampled_df)

        return sampled_df

    def predict_less_likely_samples(self, df: pd.DataFrame, n: int, temp=0.05, variaty=3):
        self.fit_sampler(df)
        return self.vae.less_likely_sample(n, temp, variaty)

    def save_state(self, path: str):
        self.vae.save_state(path)
        logger.info(f"Saved VAE state in {path}")

    def load_state(self, path: str):
        try:
            self.vae.load_state(path)

        except FileNotFoundError:
            raise FileNotFoundError("Missing file with VAE state")

        logger.info(f"Loaded VAE state from {path}")


class VanillaVAEWrapper(VAEWrapper):
    """
    Class that implements end to end train and generation of structured data by CVAE as a model.

    Attributes
    ----------
    model : CVAE
        final model that we will use to generate new data
    """
    def __init__(
            self,
            df: pd.DataFrame,
            schema: Optional[Dict],
            metadata: Dict,
            table_name: str,
            paths: Dict,
            process: str,
            main_process: str,
            batch_size: int,
            latent_dim: int = 10,
            latent_components: int = 30):

        log_level = os.getenv("LOGURU_LEVEL")

        super().__init__(
            df,
            schema,
            metadata,
            table_name,
            paths,
            process,
            main_process,
            batch_size,
            log_level
        )
        self.latent_dim = min(latent_dim, int(len(self.dataset.columns) / 2))
        self.vae = CVAE(
            self.dataset,
            batch_size=self.batch_size,
            latent_dim=latent_dim,
            latent_components=min(latent_components, latent_dim * 2),
            intermediate_dim=128,
        )
        self.vae.build_model()
        if self.process == "infer":
            self.load_state(self.paths["state_path"])



from typing import Dict, Literal, List
import json
from pathlib import Path
from dataclasses import dataclass, field

from marshmallow import (
    Schema,
    fields,
    validate,
    ValidationError,
    validates_schema,
    post_load,
)
from loguru import logger

SUPPORTED_EXCEL_EXTENSIONS = [".xls", ".xlsx"]


@dataclass
class ReportTypes:
    infer_report_types: List[str] = field(default_factory=lambda: ["accuracy", "metrics_only"])
    train_report_types: List[str] = field(init=False)
    excluded_reports: List[str] = field(default_factory=lambda: ["metrics_only"])
    full_list_of_train_report_types: List[str] = field(init=False)
    full_list_of_infer_report_types: List[str] = field(init=False)

    def __post_init__(self):
        self.train_report_types = self.infer_report_types + ["sample"]
        self.full_list_of_train_report_types = self.get_list_of_report_types("train")
        self.full_list_of_infer_report_types = self.get_list_of_report_types("infer")

    def get_list_of_report_types(self, report_type):
        """
        Get the full list of reports that should be generated
        if the parameter 'reports' sets to 'all'
        """
        report_types = (
            self.train_report_types
            if report_type == "train"
            else self.infer_report_types
        )
        return [
            report
            for report in report_types
            if report not in self.excluded_reports
        ]


class ReferenceSchema(Schema):
    table = fields.String(required=True, allow_none=False)
    columns = fields.List(fields.String(), required=True, allow_none=False)


class CaseInsensitiveString(fields.String):
    def _deserialize(self, value, attr, data, **kwargs):
        if not isinstance(value, str):
            raise ValidationError("Not a valid string.")
        return value.lower()


class KeysSchema(Schema):
    pk_types = ["PK", "UQ"]
    fk_types = ["FK"]
    type_of_keys = [*pk_types, *fk_types]
    type = fields.String(validate=validate.OneOf(type_of_keys), required=True)
    columns = fields.List(fields.String(), required=True, allow_none=False)
    joined_sample = fields.Boolean(required=False)
    references = fields.Nested(ReferenceSchema, required=False, allow_none=False)

    @validates_schema
    def validate_references(self, data, **kwargs):
        if not data["columns"]:
            raise ValidationError("The 'columns' field must not be empty")
        if data["type"] in self.fk_types and "references" not in data:
            raise ValidationError(
                f"The 'references' field is required when 'type' is "
                f"{' or '.join([f'{fk_type!r}' for fk_type in self.fk_types])}"
            )
        if data["type"] in self.fk_types and "references" in data and not data["references"]:
            raise ValidationError(
                f"The 'references' field must not be empty when 'type' is "
                f"{' or '.join([f'{fk_type!r}' for fk_type in self.fk_types])}"
            )
        if data["type"] not in self.fk_types and "references" in data:
            raise ValidationError(
                f"The 'references' field is only allowed when 'type' is "
                f"{' or '.join([f'{fk_type!r}' for fk_type in self.fk_types])}"
            )
        if data["type"] in self.fk_types and not data["references"]["columns"]:
            raise ValidationError("The 'references.columns' field must not be empty")
        if len(data["columns"]) != len(set(data["columns"])):
            raise ValidationError("The 'columns' field must contain unique values")
        if data["type"] in self.fk_types and len(data["references"]["columns"]) != len(
            set(data["references"]["columns"])
        ):
            raise ValidationError("The 'references.columns' field must contain unique values")
        if data["type"] in self.fk_types and len(data["columns"]) != len(
            data["references"].get("columns", [])
        ):
            raise ValidationError(
                "The 'columns' field must have the same length as 'references.columns'"
            )


class TrainingSettingsSchema(Schema):
    @staticmethod
    def validate_reports(x):
        if any([i in ["all", "none"] for i in x]):
            raise ValidationError(
                "The value 'all' or 'none' might not be passed in the list."
            )
        if not (
            isinstance(x, list)
            and all(
                isinstance(elem, str)
                and elem in ReportTypes().train_report_types for elem in x
            )
        ):
            raise ValidationError("Invalid value.")

    epochs = fields.Integer(validate=validate.Range(min=1), required=False)
    drop_null = fields.Boolean(required=False)
    row_limit = fields.Integer(validate=validate.Range(min=1), allow_none=True, required=False)
    batch_size = fields.Integer(validate=validate.Range(min=1), required=False)
    reports = fields.Raw(
        required=False,
        validate=validate_reports
    )


class ExtendedRestrictedTrainingSettingsSchema(TrainingSettingsSchema):
    source = fields.String(required=True, allow_none=False)
    column_types = fields.Dict(
        keys=fields.String(validate=validate.OneOf(["categorical"])),
        values=fields.List(fields.String()),
    )


class ExtendedTrainingSettingsSchema(ExtendedRestrictedTrainingSettingsSchema):
    source = fields.String(required=False, allow_none=True)


class InferSettingsSchema(Schema):
    @staticmethod
    def validate_reports(x):
        if any([i in ["all", "none"] for i in x]):
            raise ValidationError(
                "The value 'all' or 'none' might not be passed in the list."
            )
        if not (
            isinstance(x, list)
            and all(
                isinstance(elem, str)
                and elem in ReportTypes().infer_report_types for elem in x
            )
        ):
            raise ValidationError("Invalid value.")

    destination = fields.String(required=False)
    size = fields.Integer(validate=validate.Range(min=1), required=False)
    run_parallel = fields.Boolean(required=False)
    batch_size = fields.Integer(validate=validate.Range(min=1), allow_none=True, required=False)
    random_seed = fields.Integer(validate=validate.Range(min=0), allow_none=True, required=False)
    reports = fields.Raw(
        required=False,
        validate=validate_reports
    )


class CSVFormatSettingsSchema(Schema):
    sep = fields.String(required=False, allow_none=True)
    quotechar = fields.String(required=False, validate=validate.Length(equal=1))
    quoting = CaseInsensitiveString(
        required=False,
        validate=validate.OneOf(["minimal", "all", "non-numeric", "none"]),
    )
    escapechar = fields.String(required=False, validate=validate.Length(equal=1))
    encoding = fields.String(required=False, allow_none=True)
    header = fields.Raw(
        required=False,
        allow_none=True,
        validate=lambda x: isinstance(x, int)
        or (isinstance(x, str) and x == "infer")
        or (isinstance(x, list) and all(isinstance(elem, int) for elem in x)),
    )
    skiprows = fields.Raw(
        required=False,
        allow_none=True,
        validate=lambda x: isinstance(x, int)
        or (isinstance(x, list) and all(isinstance(elem, int) for elem in x)),
    )
    on_bad_lines = CaseInsensitiveString(
        required=False, validate=validate.OneOf(["error", "warn", "skip"])
    )
    engine = fields.String(
        required=False, allow_none=True, validate=validate.OneOf(["c", "python"])
    )
    na_values = fields.List(fields.String(), required=False, allow_none=True)


class ExcelFormatSettingsSchema(Schema):
    sheet_name = fields.Raw(
        required=False,
        allow_none=True,
        validate=lambda x: isinstance(x, int)
        or x is None
        or isinstance(x, str)
        or (isinstance(x, list) and all(isinstance(elem, (int, str)) for elem in x)),
    )


class GlobalSettingsSchema(Schema):
    train_settings = fields.Nested(TrainingSettingsSchema, required=False, allow_none=True)
    infer_settings = fields.Nested(InferSettingsSchema, required=False, allow_none=True)


class ConfigurationSchema(Schema):
    train_settings = fields.Nested(
        ExtendedTrainingSettingsSchema,
        required=False,
        allow_none=True
    )
    infer_settings = fields.Nested(InferSettingsSchema, required=False, allow_none=True)
    format = fields.Raw(required=False, allow_none=True)
    keys = fields.Dict(
        keys=fields.String(),
        values=fields.Nested(KeysSchema),
        required=False,
        allow_none=True,
    )

    @staticmethod
    def get_format_schema(source):
        if Path(source).suffix == ".csv":
            return CSVFormatSettingsSchema
        if Path(source).suffix in SUPPORTED_EXCEL_EXTENSIONS:
            return ExcelFormatSettingsSchema

    @post_load
    def process_format_field(self, data, **kwargs):
        train_settings = data.get("train_settings", {})
        path_to_source = train_settings.get("source") if train_settings else None
        if train_settings and path_to_source:
            format_schema = self.get_format_schema(path_to_source)
            if format_schema is not None and data.get("format") is not None:
                data["format"] = format_schema().load(data["format"])
        return data


class RestrictedConfigurationSchema(ConfigurationSchema):
    train_settings = fields.Nested(
        ExtendedRestrictedTrainingSettingsSchema,
        required=True,
        allow_none=False
    )


class ValidationSchema:
    def __init__(
        self,
        metadata: Dict,
        metadata_path: str,
        validation_source: bool,
        process: Literal["train", "infer"]
    ):
        self.metadata = metadata
        self.metadata_path = metadata_path
        self.global_schema = GlobalSettingsSchema()
        self.configuration_schema = RestrictedConfigurationSchema() \
            if validation_source and process == "train" else ConfigurationSchema()

    def validate_schema(self):
        """
        Validate the metadata file
        """
        errors = {}
        for table_name in self.metadata.keys():
            try:
                if table_name == "global":
                    self.global_schema.load(self.metadata[table_name])
                else:
                    self.configuration_schema.load(self.metadata[table_name])
            except ValidationError as err:
                errors[table_name] = err.messages
        if errors:
            message = "Validation error(s) found in the schema of the metadata"
            logger.error(message)
            for section, errors_details in errors.items():
                logger.error(
                    f'The error(s) found in - "{section}": {json.dumps(errors_details, indent=4)}'
                )
            raise ValidationError(f"{message}. The details are - {errors}")
        if not errors:
            logger.debug("The schema of the metadata is valid")



import os
import time
from threading import Thread

import streamlit as st
from streamlit_option_menu import option_menu

from syngen.streamlit_app.handlers import StreamlitHandler
from syngen.streamlit_app.utils import (
    show_data,
    get_running_status,
    set_session_state,
    cleanup_artifacts
)
from syngen import __version__


class StreamlitApp:

    @staticmethod
    def setup_ui():
        """
        Set up the UI for the Streamlit app
        """
        path_to_logo = f"{os.path.join(os.path.dirname(__file__))}/img/logo.svg"
        path_to_logo_img = f"{os.path.join(os.path.dirname(__file__))}/img/favicon.svg"
        st.set_page_config(
            page_title="SynGen UI",
            page_icon=path_to_logo_img
        )
        st.sidebar.image(path_to_logo, use_column_width=True)
        st.markdown(
            f"""
                <style>
                {"".join(open(f"{os.path.join(os.path.dirname(__file__))}/css/style.css").readlines())}
                </style>
                """, unsafe_allow_html=True
        )

        st.title(
            "SynGen UI",
            help=f"The current version of syngen library is {__version__}"
        )

    @staticmethod
    def handle_cross_icon():
        """
        Handle the behavior of disabling of the cross icon
        """
        running_status = get_running_status()

        display_status = "none" if running_status else "block"

        css = f"""
            <style>
                div[data-testid="fileDeleteBtn"] {{
                display: {display_status};
                }}
            </style>
            """
        st.markdown(css, unsafe_allow_html=True)

    @staticmethod
    def _get_streamlit_handler(epochs, size_limit, reports, uploaded_file):
        """
        Get the Streamlit handler
        """
        return StreamlitHandler(epochs, size_limit, reports, uploaded_file)

    def run_basic_page(self):
        """
        Run the basic page of the Streamlit app
        """
        set_session_state()
        uploaded_file = st.file_uploader(
            "Upload a CSV file",
            type="csv",
            accept_multiple_files=False,
            disabled=get_running_status(),
        )
        self.handle_cross_icon()
        if not uploaded_file:
            cleanup_artifacts()
            st.warning("Please upload a CSV file to proceed")
        if uploaded_file:
            show_data(uploaded_file)
            epochs = st.number_input(
                "Epochs",
                min_value=1,
                value=1,
                help="- The larger number of epochs is set the better training result is.\n"
                     "- The larger number of epochs is set the longer time for training "
                     "will be required.\n"
                     "- Actual number of epochs can be smaller that the one that was set here. "
                     "Once training stops improving the model, further training is not needed.",
                disabled=get_running_status()
            )
            size_limit = st.number_input(
                "Rows to generate",
                min_value=1,
                max_value=None,
                value=1000,
                disabled=get_running_status()
            )
            reports = st.checkbox(
                "Create an accuracy report",
                value=False,
                key="reports",
                disabled=get_running_status()
            )
            handler = self._get_streamlit_handler(epochs, size_limit, reports, uploaded_file)
            if st.button(
                    "Generate data",
                    type="primary",
                    key="gen_button",
                    disabled=get_running_status()
            ):
                runner = Thread(name="train_and_infer", target=handler.train_and_infer)
                runner.start()
                current_progress = 0
                prg = st.progress(current_progress)

                while runner.is_alive():
                    with st.expander("Logs"):
                        while True:
                            if not handler.log_queue.empty():
                                with st.code("logs", language="log"):
                                    log = handler.log_queue.get()
                                    st.text(log)
                                    current_progress, message = handler.progress_handler.info
                                    prg.progress(value=current_progress, text=message)
                            elif not handler.log_error_queue.empty():
                                runner.join()
                                break
                            elif not runner.is_alive():
                                break
                            time.sleep(0.001)
                if not handler.log_error_queue.empty() and not runner.is_alive():
                    st.exception(handler.log_error_queue.get())
                elif handler.log_queue.empty() and not runner.is_alive():
                    prg.progress(100)
                    st.success("Data generation completed")

            with st.container():
                handler.generate_buttons()

    def run(self):
        """
        Run the Streamlit app
        """
        self.setup_ui()
        with st.sidebar:
            selected = option_menu("", ["Basic"],
                                   icons=["'play'"],
                                   default_index=0,
                                   menu_icon=None,
                                   styles={
                                       "container": {"font-family": "Open Sans"}
                                   }
                                   )
        if selected == "Basic":
            self.run_basic_page()


if __name__ == "__main__":
    StreamlitApp().run()



from syngen.ml.data_loaders.data_loaders import (  # noqa: F401;
    CSVLoader,
    AvroLoader,
    DataLoader,
    MetadataLoader,
    YAMLLoader,
    BaseDataLoader,
    BinaryLoader,
    ExcelLoader,
    DataEncryptor
)
from syngen.ml.data_loaders.dataframe_fetcher import DataFrameFetcher  # noqa: F401
